# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/Importing.ipynb (unless otherwise specified).

__all__ = ['read_file', 'extract_rawfile_unique_values', 'import_spectronaut_data', 'import_maxquant_data',
           'convert_ap_mq_mod', 'import_alphapept_data', 'convert_diann_mq_mod', 'import_diann_data',
           'convert_fragpipe_mq_mod', 'import_fragpipe_data', 'import_data']

# Cell
import os
import pandas as pd

def read_file(
    file: str,
    column_names: list
) -> pd.DataFrame:
    """Load a specified columns of the file as a pandas dataframe.

    Args:
        file (str): The name of a file.
        column_names (list): The list of three columns that should be extracted from the file.

    Raises:
        NotImplementedError: if a specified file has not a .csv, .txt or .tsv extension.
        ValueError: if any of the specified columns is not in the file.

    Returns:
        pd.DataFrame: A pandas dataframe with all the data stored in the specified columns.
    """
    file_ext = os.path.splitext(file)[-1]
    if file_ext=='.csv':
        sep=','
    elif file_ext=='.tsv':
        sep='\t'
    elif file_ext=='.txt':
        sep='\t'
    else:
        raise NotImplementedError("The selected filetype isn't supported. Please specify a file with a .csv, .txt or .tsv extension.")
    with open(file) as filelines:
        i = 0
        pos = 0
        for l in filelines:
            i += 1
            l = l.split(sep)
            try:
                raw = l.index(column_names[0])
                prot = l.index(column_names[1])
                seq = l.index(column_names[2])
            except:
                raise ValueError('The list of specified column names cannot be extracted from the file.')
            if i>0:
                break

    with open(file) as filelines:
        raws = []
        prots = []
        seqs = []
        for l in filelines:
            l = l.split(sep)
            raws.append(l[raw])
            prots.append(l[prot])
            seqs.append(l[seq])

    res = pd.DataFrame({column_names[0]:raws[1:],
             column_names[1]:prots[1:],
             column_names[2]:seqs[1:]})

    return res


def extract_rawfile_unique_values(
    file: str
) -> list:
    """Extract the unique raw file names from "R.FileName" (Spectronaut output), "Raw file" (MaxQuant output),
    "shortname" (AlphaPept output) or "Run" (DIA-NN output) column or from the "Spectral Count" column from the
    combined_peptide.tsv file without modifications for the FragPipe.

    Args:
        file (str): The name of a file.

    Raises:
        ValueError: if a column with the unique raw file names is not in the file.

    Returns:
        list: A sorted list of unique raw file names from the file.
    """
    file_ext = os.path.splitext(file)[-1]
    if file_ext == '.csv':
        sep = ','
    elif file_ext in ['.tsv', '.txt']:
        sep = '\t'

    with open(file) as filelines:
        i = 0
        filename_col_index = None
        filename_data = []

        for l in filelines:
            l = l.split(sep)
            # just do it for the first line
            if i == 0:
                for col in ['R.FileName', 'Raw file', 'Run', 'shortname']:
                    if col in l:
                        filename_col_index = l.index(col)
                        break
                if not isinstance(filename_col_index, int):
                    # to check the case with the FragPipe peptide.tsv file when we don't have the info about the experiment name
                    if ("Assigned Modifications" in "".join(l)) and ("Protein ID" in "".join(l)) and ("Peptide" in "".join(l)):
                        return []
                    # to check the case with the FragPipe combined_peptide.tsv file when the experiment name is included in the "Spectral Count" column
                    elif ("Sequence" in "".join(l)) and ("Assigned Modifications" in "".join(l)) and ("Protein ID" in "".join(l)):
                        return sorted(list(set([col.replace('_', '').replace(' Spectral Count', '') for col in l if 'Spectral Count' in col])))
                    else:
                        raise ValueError('A column with the raw file names is not in the file.')
            else:
                filename_data.append(l[filename_col_index])
            i += 1

        unique_filenames = set(filename_data)

    sorted_unique_filenames = sorted(list(unique_filenames))
    return sorted_unique_filenames

# Cell
import pandas as pd
import re
from typing import Union

def import_spectronaut_data(
    file: str,
    sample: Union[str, list, None] = None
) -> pd.DataFrame:
    """Import peptide level data from Spectronaut.

    Args:
        file (str): The name of a file.
        sample (Union[str, list, None]): The unique raw file name(s) to filter the original file. Defaults to None. In this case data for all raw files will be extracted.

    Returns:
        pd.DataFrame: A pandas dataframe containing information about: all_protein_ids (str), modified_sequence (str), naked_sequence (str)
    """
    spectronaut_columns = ["PEP.AllOccurringProteinAccessions","EG.ModifiedSequence","R.FileName"]

    data = read_file(file, spectronaut_columns)

    if sample:
        if isinstance(sample, list):
            data_sub = data[data["R.FileName"].isin(sample)]
            data_sub = data_sub[["PEP.AllOccurringProteinAccessions","EG.ModifiedSequence"]]
        elif isinstance(sample, str):
            data_sub = data[data["R.FileName"] == sample]
            data_sub = data_sub[["PEP.AllOccurringProteinAccessions","EG.ModifiedSequence"]]
    else:
        data_sub = data[["PEP.AllOccurringProteinAccessions","EG.ModifiedSequence"]]

    # get modified sequence
    mod_seq = data_sub.apply(lambda row: re.sub('_','',row["EG.ModifiedSequence"]), axis=1)
    data_sub = data_sub.assign(modified_sequence=mod_seq.values)
    # get naked sequence
    nak_seq = data_sub.apply(lambda row: re.sub(r'\[.*?\]','',row["modified_sequence"]), axis=1)
    data_sub = data_sub.assign(naked_sequence=nak_seq.values)
    data_sub = data_sub.rename(columns={"PEP.AllOccurringProteinAccessions": "all_protein_ids"})
    input_data = data_sub[["all_protein_ids","modified_sequence","naked_sequence"]]
    input_data = input_data.dropna()
    input_data = input_data.drop_duplicates().reset_index(drop=True)
    return input_data

# Cell
import pandas as pd
from typing import Union
import re

def import_maxquant_data(
    file: str,
    sample: Union[str, list, None] = None
) -> pd.DataFrame:
    """Import peptide level data from MaxQuant.

    Args:
        file (str): The name of a file.
        sample (Union[str, list, None]): The unique raw file name(s) to filter the original file. Defaults to None. In this case data for all raw files will be extracted.

    Returns:
        pd.DataFrame: A pandas dataframe containing information about: all_protein_ids (str), modified_sequence (str), naked_sequence (str)
    """
    mq_columns = ["Proteins","Modified sequence","Raw file"]

    data = read_file(file, mq_columns)

    if sample:
        if isinstance(sample, list):
            data_sub = data[data["Raw file"].isin(sample)]
            data_sub = data_sub[["Proteins","Modified sequence"]]
        elif isinstance(sample, str):
            data_sub = data[data["Raw file"] == sample]
            data_sub = data_sub[["Proteins","Modified sequence"]]
    else:
        data_sub = data[["Proteins","Modified sequence"]]

    # get modified sequence
    mod_seq = data_sub.apply(lambda row: re.sub('_','',row["Modified sequence"]), axis=1)
    data_sub = data_sub.assign(modified_sequence=mod_seq.values)

    # replace outer () with []
    mod_seq_replaced = data_sub.apply(lambda row: re.sub(r'\((.*?\(.*?\))\)',r'[\1]',row["modified_sequence"]), axis=1)
    data_sub = data_sub.assign(modified_sequence=mod_seq_replaced.values)

    # get naked sequence
    nak_seq = data_sub.apply(lambda row: re.sub(r'\[.*?\]','',row["modified_sequence"]), axis=1)
    data_sub = data_sub.assign(naked_sequence=nak_seq.values)
    data_sub = data_sub.rename(columns={"Proteins": "all_protein_ids"})
    input_data = data_sub[["all_protein_ids","modified_sequence","naked_sequence"]]
    input_data = input_data.dropna() # remove missing values
    input_data = input_data.drop_duplicates().reset_index(drop=True)
    return input_data

# Cell
import re

def convert_ap_mq_mod(
    sequence:str
) -> str:
    """Convert AlphaPept style modifications into MaxQuant style modifications.

    Args:
        sequence (str): The peptide sequence with modification in an AlphaPept style.

    Returns:
        str: The peptide sequence with modification in a similar to MaxQuant style.
    """
    # TODO: add more AP modifications
    modif_convers_dict = {
        'ox': '[Oxidation ({})]',
        'a': '[Acetyl ({})]',
        'am': '[Amidated ({})]',
        'deam': '[Deamidation ({})]',
        'p': '[Phospho ({})]',
        'pg': '[{}->pyro-Glu]',
        'c': '[Cys-Cys]'
    }
    mods = re.findall('[a-z0-9]+', sequence)
    if mods:
        for mod in mods:
            posit = re.search(mod, sequence)
            i = posit.start()
            if i == 0 and mod == 'a':
                add_aa = 'N-term'
            elif posit.end() == len(sequence) - 1 and mod == 'am':
                add_aa = sequence[posit.end()]
                sequence = sequence.replace(mod + add_aa, add_aa + mod, 1)
                add_aa = 'C-term'
            else:
                add_aa = sequence[posit.end()]
                sequence = sequence.replace(mod + add_aa, add_aa + mod, 1)

            if mod == 'ox':
                if add_aa == 'M':
                    add_aa = 'M'
                elif add_aa in 'MP':
                    add_aa = 'MP'
            elif mod == 'deam':
                if add_aa in 'NQ':
                    add_aa = 'NQ'
            elif mod == 'p':
                if add_aa in 'STY':
                    add_aa = 'STY'
            elif mod == 'pg':
                if add_aa == 'E':
                    add_aa = 'Glu'
                elif add_aa == 'Q':
                    add_aa = 'Gln'

            if mod in modif_convers_dict.keys():
                sequence = sequence.replace(mod, modif_convers_dict.get(mod).format(add_aa), 1)
    return sequence

# Cell
import pandas as pd
from typing import Union

def import_alphapept_data(
    file: str,
    sample: Union[str, list, None] = None
) -> pd.DataFrame:
    """Import peptide level data from AlphaPept.

    Args:
        file (str): The name of a file.
        sample (Union[str, list, None]): The unique raw file name(s) to filter the original file. Defaults to None. In this case data for all raw files will be extracted.

    Returns:
        pd.DataFrame: A pandas dataframe containing information about: all_protein_ids (str), modified_sequence (str), naked_sequence (str)
    """
    ap_columns = ["protein_group", "sequence", "shortname"]

    data = pd.read_csv(file, usecols=ap_columns)
    # TODO: add later the file reading using read_file function. For now it doesn't work for the protein groups that should be split later

    if sample:
        if isinstance(sample, list):
            data_sub = data[data["shortname"].isin(sample)]
            data_sub = data_sub[["protein_group", "sequence"]]
        elif isinstance(sample, str):
            data_sub = data[data["shortname"] == sample]
            data_sub = data_sub[["protein_group", "sequence"]]
    else:
        data_sub = data[["protein_group", "sequence"]]

    data_sub = data_sub[~data_sub.sequence.str.contains('_decoy')]

    # get modified sequence
    modif_seq = data_sub.apply(lambda row: convert_ap_mq_mod(row.sequence), axis=1)
    data_sub['modified_sequence'] = modif_seq.values

    # get a list of proteins_id
    proteins = data_sub.apply(lambda row: ";".join([_.split('|')[1] for _ in row.protein_group.split(',')]), axis=1)
    data_sub['all_protein_ids'] = proteins.values

    # get naked sequence
    nak_seq = data_sub.apply(lambda row: ''.join([_ for _ in row.sequence if _.isupper()]), axis=1)
    data_sub['naked_sequence'] = nak_seq.values

    input_data = data_sub[["all_protein_ids", "modified_sequence", "naked_sequence"]]
    input_data = input_data.dropna() # remove missing values
    input_data = input_data.drop_duplicates().reset_index(drop=True)
    return input_data

# Cell
import re

def convert_diann_mq_mod(
    sequence:str
) -> str:
    """Convert DIA-NN style modifications into MaxQuant style modifications.

    Args:
        sequence (str): The peptide sequence with modification in an AlphaPept style.

    Returns:
        str: The peptide sequence with modification in a similar to DIA-NN style.
    """

    modif_convers_dict = {
        '(UniMod:1)': '[Acetyl ({})]',
        '(UniMod:2)': '[Amidated ({})]',
        '(UniMod:4)': '[Carbamidomethyl ({})]',
        '(UniMod:5)': '[Carbamyl ({})]',
        '(UniMod:7)': '[Deamidation ({})]',
        '(UniMod:21)': '[Phospho ({})]',
        '(UniMod:23)': '[Dehydrated ({})]',
        '(UniMod:26)': '[Pyro-carbamidomethyl ({})]',
        '(UniMod:27)': '[Glu->pyro-Glu]',
        '(UniMod:28)': '[Gln->pyro-Glu]',
        '(UniMod:30)': '[Cation:Na ({})]',
        '(UniMod:34)': '[Methyl ({})]',
        '(UniMod:35)': '[Oxidation ({})]',
        '(UniMod:36)': '[Dimethyl ({})]',
        '(UniMod:37)': '[Trimethyl ({})]',
        '(UniMod:40)': '[Sulfo ({})]',
        '(UniMod:55)': '[Cys-Cys]',
        '(UniMod:121)': '[GlyGly ({})]',
        '(UniMod:254)': '[Delta:H(2)C(2) ({})]',
        '(UniMod:312)': '[Cysteinyl]',
        '(UniMod:345)': '[Trioxidation ({})]',
        '(UniMod:408)': '[Hydroxyproline]',
        '(UniMod:425)': '[Dioxidation ({})]',
        '(UniMod:526)': '[Dethiomethyl ({})]',
        '(UniMod:877)': '[QQTGG ({})]',
    }
    mods = re.findall('\(UniMod:\d+\)', sequence)
    if mods:
        for mod in mods:
            posit = re.search('\(UniMod:\d+\)', sequence)
            i = posit.start()

            if i == 0:
                add_aa = 'N-term'
            elif posit.end() == len(sequence):
                add_aa = 'C-term'
            else:
                add_aa = sequence[i-1]

            if mod == '(UniMod:7)':
                if add_aa in 'NQ':
                    add_aa = 'NQ'
            elif mod == '(UniMod:21)':
                if add_aa in 'STY':
                    add_aa = 'STY'
            elif mod == '(UniMod:23)':
                if add_aa in 'ST':
                    add_aa = 'ST'
            elif mod == '(UniMod:30)':
                if add_aa in 'DE':
                    add_aa = 'DE'
            elif mod == '(UniMod:34)':
                if add_aa in 'KR':
                    add_aa = 'KR'
            elif mod == '(UniMod:36)':
                if add_aa in 'KR':
                    add_aa = 'KR'
            elif mod == '(UniMod:40)':
                if add_aa in 'STY':
                    add_aa = 'STY'
            elif mod == '(UniMod:425)':
                if add_aa in 'MW':
                    add_aa = 'MW'

            if mod in modif_convers_dict.keys():
                sequence = sequence.replace(mod, modif_convers_dict.get(mod).format(add_aa), 1)

    return sequence

# Cell
import pandas as pd
from typing import Union

def import_diann_data(
    file: str,
    sample: Union[str, list, None] = None
) -> pd.DataFrame:
    """Import peptide level data from DIA-NN.

    Args:
        file (str): The name of a file.
        sample (Union[str, list, None]): The unique raw file name(s) to filter the original file. Defaults to None. In this case data for all raw files will be extracted.

    Returns:
        pd.DataFrame: A pandas dataframe containing information about: all_protein_ids (str), modified_sequence (str), naked_sequence (str)
    """
    diann_columns = ["Protein.Ids", "Modified.Sequence", "Run"]

    data = read_file(file, diann_columns)

    if sample:
        if isinstance(sample, list):
            data_sub = data[data["Run"].isin(sample)]
            data_sub = data_sub[["Protein.Ids", "Modified.Sequence"]]
        elif isinstance(sample, str):
            data_sub = data[data["Run"] == sample]
            data_sub = data_sub[["Protein.Ids", "Modified.Sequence"]]
    else:
        data_sub = data[["Protein.Ids", "Modified.Sequence"]]

    # get a list of proteins_id
    data_sub = data_sub.rename(columns={"Protein.Ids": "all_protein_ids"})

    # get modified sequence
    modif_seq = data_sub.apply(lambda row: convert_diann_mq_mod(row["Modified.Sequence"]), axis=1)
    data_sub['modified_sequence'] = modif_seq.values

    # get naked sequence
    nak_seq = data_sub.apply(lambda row: re.sub(r'\[.*?\]', '', row["modified_sequence"]), axis=1)
    data_sub = data_sub.assign(naked_sequence = nak_seq.values)

    input_data = data_sub[["all_protein_ids", "modified_sequence", "naked_sequence"]]
    input_data = input_data.dropna() # remove missing values
    input_data = input_data.drop_duplicates().reset_index(drop=True)
    return input_data

# Cell
import re

def convert_fragpipe_mq_mod(
    sequence:str,
    assigned_modifications: str
) -> str:
    """Convert FragPipe style modifications into MaxQuant style modifications.

    Args:
        sequence (str): The peptide sequence with modification.
        assigned_modifications (str): The string of assigned modifications separated by comma.

    Returns:
        str: The peptide sequence with modification in a similar to DIA-NN style.
    """
    modif_convers_dict = {
        42.0106: '[Acetyl ({})]',
        -0.9840: '[Amidated ({})]',
        57.0215: '[Carbamidomethyl ({})]',
        43.0058: '[Carbamyl ({})]',
        0.9840: '[Deamidation ({})]',
        79.9663: '[Phospho ({})]',
        -18.0106: ['[Dehydrated ({})]', '[Glu->pyro-Glu]'],
        39.9949: '[Pyro-carbamidomethyl ({})]',
        -17.0265: '[Gln->pyro-Glu]',
        21.9819: '[Cation:Na ({})]',
        14.0157: '[Methyl ({})]',
        15.9949: '[Oxidation ({})]',
        28.0313: '[Dimethyl ({})]',
        42.047: '[Trimethyl ({})]',
        79.9568: '[Sulfo ({})]',
        305.0682: '[Cys-Cys]',
        114.0429: '[GlyGly ({})]',
        26.0157: '[Delta:H(2)C(2) ({})]',
        119.0041: '[Cysteinyl]',
        47.9847: '[Trioxidation ({})]',
        148.0372: '[Hydroxyproline]',
        31.9898: '[Dioxidation ({})]',
        -48.0034: '[Dethiomethyl ({})]',
        599.2663: '[QQTGG ({})]',
    }

    if assigned_modifications:
        modifs_posit = [''] * (len(sequence) + 1)
        for mod in assigned_modifications.split(','):
            mod = mod.strip()
            data = mod.replace(')', '').replace('"', '').split('(')
            mod_pos, mod_mass = data[0], float(data[1])
            if mod_pos == 'N-term':
                posit = 0
                add_aa = 'N-term'
            elif mod_pos == 'C-term':
                posit = -1
                add_aa = 'C-term'
            else:
                posit = int(mod_pos[:-1])
                add_aa = mod_pos[-1]
                if mod_mass == 0.9840:
                    if add_aa in 'NQ':
                        add_aa = 'NQ'
                elif mod_mass == 79.9663:
                    if add_aa in 'STY':
                        add_aa = 'STY'
                elif mod_mass == 21.9819:
                    if add_aa in 'DE':
                        add_aa = 'DE'
                elif mod_mass == 14.0157:
                    if add_aa in 'KR':
                        add_aa = 'KR'
                elif mod_mass == 28.0313:
                    if add_aa in 'KR':
                        add_aa = 'KR'
                elif mod_mass == 79.9568:
                    if add_aa in 'STY':
                        add_aa = 'STY'
                elif mod_mass == 31.9898:
                    if add_aa in 'MW':
                        add_aa = 'MW'
            if mod_mass == -18.0106:
                if add_aa == 'E':
                    modifs_posit[posit] = modif_convers_dict[mod_mass][1].format(add_aa)
                else:
                    if add_aa in 'ST':
                        add_aa = 'ST'
                    modifs_posit[posit] = modif_convers_dict[mod_mass][0].format(add_aa)
            else:
                modifs_posit[posit] = modif_convers_dict[mod_mass].format(add_aa)

        modif_sequence = ''.join(["".join(i) for i in zip(' '+ sequence, modifs_posit)]).strip()
        return modif_sequence

    else:
        return sequence

# Cell
import pandas as pd
from typing import Union

def import_fragpipe_data(
    file: str,
    sample: Union[str, list, None] = None
) -> pd.DataFrame:
    """Import peptide level data from FragPipe/MSFragger.

    Args:
        file (str): The name of a file.
        sample (Union[str, list, None]): The unique raw file name(s) to filter the original file. Defaults to None. In this case data for all raw files will be extracted.

    Returns:
        pd.DataFrame: A pandas dataframe containing information about: all_protein_ids (str), modified_sequence (str), naked_sequence (str)
    """
    file_ext = os.path.splitext(file)[-1]
    if file_ext=='.csv':
        sep=','
    elif file_ext=='.tsv':
        sep='\t'
    elif file_ext=='.txt':
        sep='\t'
    if sample:
        if isinstance(sample, list):
            column_names = [each + ' Spectral Count' for each in sample]
            combined_fragpipe_columns = ["Sequence", "Protein ID"] + column_names
            data = pd.read_csv(file, sep=sep, low_memory=False, usecols=combined_fragpipe_columns)
            selected_indices = []
            for column_name in column_names:
                selected_indices.extend(data[data[column_name] > 0].index.tolist())
            data_sub = data.iloc[list(set(selected_indices))]
            data_sub = data_sub[["Sequence", "Protein ID"]]
        elif isinstance(sample, str):
            column_name = sample + ' Spectral Count'
            combined_fragpipe_columns = ["Sequence", "Protein ID", column_name]
            data = pd.read_csv(file, sep=sep, low_memory=False, usecols=combined_fragpipe_columns)
            selected_indices = data[data[column_name] > 0].index.tolist()
            data_sub = data.iloc[selected_indices]
            data_sub = data_sub[["Sequence", "Protein ID"]]

        # rename columns into all_proteins_id and naked sequence
        data_sub = data_sub.rename(columns={"Protein ID": "all_protein_ids", "Sequence": "naked_sequence"})
        data_sub['modified_sequence'] = data_sub.naked_sequence

    else:
        try:
            combined_fragpipe_columns = ["Sequence", "Protein ID"]
            data_sub = pd.read_csv(file, sep=sep, low_memory=False, usecols=combined_fragpipe_columns)

            # rename columns into all_proteins_id and naked sequence
            data_sub = data_sub.rename(columns={"Protein ID": "all_protein_ids", "Sequence": "naked_sequence"})
            data_sub['modified_sequence'] = data_sub.naked_sequence
        except:
            fragpipe_columns = ["Protein ID", "Peptide", "Assigned Modifications"]
            data = read_file(file, fragpipe_columns)
            data_sub = data[["Protein ID", "Peptide", "Assigned Modifications"]]

            # get modified sequence
            modif_seq = data_sub.apply(lambda row: convert_fragpipe_mq_mod(row["Peptide"], row["Assigned Modifications"]), axis=1)
            data_sub['modified_sequence'] = modif_seq.values

            # rename columns into all_proteins_id and naked sequence
            data_sub = data_sub.rename(columns={"Protein ID": "all_protein_ids", "Peptide": "naked_sequence"})

    input_data = data_sub[["all_protein_ids", "modified_sequence", "naked_sequence"]]
    input_data = input_data.dropna() # remove missing values
    input_data = input_data.drop_duplicates().reset_index(drop=True)
    return input_data

# Cell
import pandas as pd
import re
from io import StringIO
import os

def import_data(
    file :str,
    sample: Union[str, list, None] = None,
    verbose: bool = True,
    dashboard: bool = False
) -> pd.DataFrame:
    """Import peptide level data. Depending on available columns in the provided file, the function calls other specific functions for each tool.

    Args:
        file (str): The name of a file.
        sample (Union[str, list, None]): The unique raw file name(s) to filter the original file. Defaults to None. In this case data for all raw files will be extracted.
        verbose (bool): if True, print the type of input that is used. Defaults to True.
        dashboard (bool): If True, the function is used for the dashboard output (StringIO object). Defaults to False.

    Raises:
        TypeError: If the input data format is unknown.

    Returns:
        pd.DataFrame: A pandas dataframe containing information about: all_protein_ids (str), modified_sequence (str), naked_sequence (str)
    """
    if dashboard:
        file = StringIO(str(file, "utf-8"))

    file_ext = os.path.splitext(file)[-1]
    if file_ext=='.csv':
        sep=','
    elif file_ext=='.tsv':
        sep='\t'
    elif file_ext=='.txt':
        sep='\t'

    with open(file) as filelines:
        i = 0
        pos = 0
        for l in filelines:
            i += 1
            l = l.strip().split(sep)
            if i>0:
                break

        uploaded_data_columns = set(l)
        input_info = file

    if set(["Proteins","Modified sequence","Raw file"]).issubset(uploaded_data_columns):
        if verbose:
            print("Import MaxQuant output")
        data = import_maxquant_data(input_info, sample=sample)
    elif set(["PEP.AllOccurringProteinAccessions","EG.ModifiedSequence","R.FileName"]).issubset(uploaded_data_columns):
        if verbose:
            print("Import Spectronaut output")
        data = import_spectronaut_data(input_info, sample=sample)
    elif set(["protein_group", "sequence", "shortname"]).issubset(uploaded_data_columns):
        if verbose:
            print("Import AlphaPept output")
        data = import_alphapept_data(input_info, sample=sample)
    elif set(["Protein.Ids", "Modified.Sequence", "Run"]).issubset(uploaded_data_columns):
        if verbose:
            print("Import DIA-NN output")
        data = import_diann_data(input_info, sample=sample)
    elif set(["Protein ID", "Assigned Modifications"]).issubset(uploaded_data_columns):
        if verbose:
            print("Import FragPipe output")
        data = import_fragpipe_data(input_info, sample=sample)
    else:
        raise TypeError(f'Input data format for {file} not known.')
    return data


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"read_file": "Importing.ipynb",
         "extract_rawfile_unique_values": "Importing.ipynb",
         "import_spectronaut_data": "Importing.ipynb",
         "import_maxquant_data": "Importing.ipynb",
         "convert_ap_mq_mod": "Importing.ipynb",
         "import_alphapept_data": "Importing.ipynb",
         "convert_diann_mq_mod": "Importing.ipynb",
         "import_diann_data": "Importing.ipynb",
         "convert_fragpipe_mq_mod": "Importing.ipynb",
         "import_fragpipe_data": "Importing.ipynb",
         "import_data": "Importing.ipynb",
         "extract_uniprot_id": "Preprocessing.ipynb",
         "expand_protein_ids": "Preprocessing.ipynb",
         "pep_position_helper": "Preprocessing.ipynb",
         "get_peptide_position": "Preprocessing.ipynb",
         "get_ptm_sites": "Preprocessing.ipynb",
         "get_modifications": "Preprocessing.ipynb",
         "format_input_data": "Preprocessing.ipynb",
         "format_uniprot_annotation": "SequencePlot.ipynb",
         "ptm_shape_dict": "SequencePlot.ipynb",
         "get_plot_data": "SequencePlot.ipynb",
         "plot_single_peptide_traces": "SequencePlot.ipynb",
         "custom_color_palettes": "SequencePlot.ipynb",
         "uniprot_color_dict": "SequencePlot.ipynb",
         "aa_color_dict": "SequencePlot.ipynb",
         "get_quality_category": "SequencePlot.ipynb",
         "get_exposure_category": "SequencePlot.ipynb",
         "get_alphafold_annotation": "SequencePlot.ipynb",
         "plot_peptide_traces": "SequencePlot.ipynb",
         "extract_annotation": "SequencePlot.ipynb",
         "manipulate_cif": "SequencePlot.ipynb",
         "adjust_html": "SequencePlot.ipynb",
         "get_ms_concensus": "SequencePlot.ipynb",
         "plot_3d_structure": "SequencePlot.ipynb",
         "create_pdf_report": "SequencePlot.ipynb",
         "extract_note": "Uniprot_integration.ipynb",
         "extract_note_end": "Uniprot_integration.ipynb",
         "resolve_unclear_position": "Uniprot_integration.ipynb",
         "extract_positions": "Uniprot_integration.ipynb",
         "preprocess_uniprot": "Uniprot_integration.ipynb",
         "uniprot_feature_dict": "Uniprot_integration.ipynb",
         "all_organisms": "organisms_data.ipynb",
         "import_fasta": "organisms_data.ipynb",
         "import_uniprot_annotation": "organisms_data.ipynb",
         "protease_dict": "proteolytic_cleavage.ipynb",
         "get_cleavage_sites": "proteolytic_cleavage.ipynb"}

modules = ["importing.py",
           "preprocessing.py",
           "sequenceplot.py",
           "uniprot_integration.py",
           "organisms_data.py",
           "proteolytic_cleavage.py"]

doc_url = "https://mannlabs.github.io/alphamap/"

git_url = "https://github.com/mannlabs/alphamap/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/organisms_data.ipynb (unless otherwise specified).

__all__ = ['all_organisms', 'import_fasta', 'import_uniprot_annotation']

# Cell
all_organisms = {
    'Human': {'fasta_name': 'human.fasta',
               'uniprot_name': 'preprocessed_uniprot_human.csv'
              },
    'Mouse': {'fasta_name': 'mouse.fasta',
               'uniprot_name': 'preprocessed_uniprot_mouse.csv'
              },
    'Rat': {'fasta_name': 'rat_[10116].fasta',
               'uniprot_name': 'preprocessed_uniprot_rat.csv'
              },
    'Cow': {'fasta_name': 'bovine_[9913].fasta',
               'uniprot_name': 'preprocessed_uniprot_bovine.csv'
              },
    'Zebrafish': {'fasta_name': 'zebrafish_[7955].fasta',
               'uniprot_name': 'preprocessed_uniprot_zebrafish.csv'
              },
    'Drosophila': {'fasta_name': 'drosophila_[7227].fasta',
               'uniprot_name': 'preprocessed_uniprot_drosophila.csv'
              },
    'Caenorhabditis elegans': {'fasta_name': 'celegans_[6239].fasta',
               'uniprot_name': 'preprocessed_uniprot_human.csv'
              },
    'Slime mold': {'fasta_name': 'slimemold_[44689].fasta',
               'uniprot_name': 'preprocessed_uniprot_slimemold.csv'
              },
    'Arabidopsis thaliana': {'fasta_name': 'arabidopsis_thaliana.fasta',
               'uniprot_name': 'preprocessed_uniprot_arabidopsis.csv'
              },
    'Rice': {'fasta_name': 'rice_[39947].fasta',
               'uniprot_name': 'preprocessed_uniprot_rice.csv'
              },
    'Escherichia coli': {'fasta_name': 'ecoli_[83333].fasta',
               'uniprot_name': 'preprocessed_uniprot_ecoli.csv'
              },
    'Bacillus subtilis': {'fasta_name': 'bsubtilis_[224308].fasta',
               'uniprot_name': 'preprocessed_uniprot_bsubtilis.csv'
              },
    'Saccharomyces cerevisiae': {'fasta_name': 'yeast_[559292].fasta',
               'uniprot_name': 'preprocessed_uniprot_yeast.csv'
              },
    'SARS-CoV': {'fasta_name': 'cov.fasta',
               'uniprot_name': 'preprocessed_uniprot_cov.csv'
              },
    'SARS-CoV2': {'fasta_name': 'cov2.fasta',
               'uniprot_name': 'preprocessed_uniprot_cov2.csv'
              }
}

# Cell
import os
import urllib.request
import shutil
import imp
from pyteomics import fasta
def import_fasta(organism: str):
    """
    Import fasta file for the selected organism.
    This downloads the file from github if not present.

    Args:
        organism (str): Organism for which the fasta file should be imported.
    Returns:
        fasta: Fasta file imported by pyteomics 'fasta.IndexedUniProt' for the selected organism.
    """
    if not organism in all_organisms.keys():
        raise ValueError(f"Organism {organism} is not available. Please select one of the following: {list(all_organisms.keys())}")


    BASE_PATH = imp.find_module("alphamap")[1] #os.path.abspath('')
    INI_PATH = os.path.join(BASE_PATH, '..')
    FUNCT_PATH = os.path.join(INI_PATH, 'alphamap')
    DATA_PATH = os.path.join(FUNCT_PATH, 'data')

    fasta_name = all_organisms[organism]['fasta_name']

    if not os.path.exists(os.path.join(DATA_PATH, fasta_name)):
        print(f"The fasta file for {organism} is downloaded from github.")
        github_url_data_folder = 'https://github.com/MannLabs/alphamap/blob/master/alphamap/data/'

        github_file = os.path.join(
            github_url_data_folder,
            os.path.basename(os.path.join(DATA_PATH, fasta_name))) + '/?raw=true'

        with urllib.request.urlopen(github_file) as response, open(os.path.join(DATA_PATH, fasta_name), 'wb') as out_file:
            shutil.copyfileobj(response, out_file)

    fasta_file = fasta.IndexedUniProt(os.path.join(DATA_PATH, fasta_name))

    return fasta_file

# Cell
import os
import urllib.request
import shutil
import imp
import pandas as pd
def import_uniprot_annotation(organism: str):
    """
    Import uniprot annotation file for the selected organism.
    This downloads the file from github if not present.

    Args:
        organism (str): Organism for which the uniprot annotation should be imported.
    Returns:
        pd.DataFrame: Dataframe with the uniprot annotations for the selected organism.

    """
    if not organism in all_organisms.keys():
        raise ValueError(f"Organism {organism} is not available. Please select one of the following: {list(all_organisms.keys())}")


    BASE_PATH = imp.find_module("alphamap")[1] #os.path.abspath('')
    INI_PATH = os.path.join(BASE_PATH, '..')
    FUNCT_PATH = os.path.join(INI_PATH, 'alphamap')
    DATA_PATH = os.path.join(FUNCT_PATH, 'data')

    uniprot_name = all_organisms[organism]['uniprot_name']

    if not os.path.exists(os.path.join(DATA_PATH, uniprot_name)):
        print(f"The uniprot annotation file for {organism} is downloaded from github.")
        github_url_data_folder = 'https://github.com/MannLabs/alphamap/blob/master/alphamap/data/'

        github_file = os.path.join(
            github_url_data_folder,
            os.path.basename(os.path.join(DATA_PATH, uniprot_name))) + '/?raw=true'

        with urllib.request.urlopen(github_file) as response, open(os.path.join(DATA_PATH, uniprot_name), 'wb') as out_file:
            shutil.copyfileobj(response, out_file)

    uniprot_file = pd.read_csv(os.path.join(DATA_PATH, uniprot_name))

    return uniprot_file


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/SequencePlot.ipynb (unless otherwise specified).

__all__ = ['format_uniprot_annotation', 'ptm_shape_dict', 'get_plot_data', 'plot_single_peptide_traces',
           'custom_color_palettes', 'uniprot_color_dict', 'aa_color_dict', 'get_quality_category',
           'get_exposure_category', 'get_alphafold_annotation', 'plot_peptide_traces', 'extract_annotation',
           'manipulate_cif', 'adjust_html', 'get_ms_concensus', 'plot_3d_structure', 'create_pdf_report']

# Cell
import os
import tempfile

# Cell
import pandas as pd

def format_uniprot_annotation(uniprot_ann: pd.DataFrame, uniprot_feature_dict: dict):
    """
    Function to format uniprot annotation for plotting.

    Args:
        uniprot_ann (pd.DataFrame): Formatted uniprot annotations from alphamap.
        uniprot_feature_dict (dict): Uniprot feature dictionary defined by alphamap.
    Returns:
        pd.DataFrame: Uniprot annotation with a combined structure entry for helix, strand and turn.

    """
    uniprot = uniprot_ann.copy(deep=True)
    uniprot.loc[uniprot.feature == "HELIX", "note"] = "Helix"
    uniprot.loc[uniprot.feature == "STRAND", "note"] = "Beta strand"
    uniprot.loc[uniprot.feature == "TURN", "note"] = "Turn"
    uniprot.loc[uniprot.feature.isin(["HELIX","STRAND","TURN"]), "feature"] = "STRUCTURE"

    uniprot_feature_dict_rev = {v: k for k, v in uniprot_feature_dict.items()}

    uniprot['annotation'] = uniprot['note']
    uniprot.loc[uniprot['annotation'].isnull(), 'annotation'] = uniprot['feature']
    uniprot = uniprot.replace({"annotation": uniprot_feature_dict_rev})
    return uniprot

# Cell

# The PTM shape dictionary defines the shapes for each modification, so they are consistent across datasets and analyses.

ptm_shape_dict = {
    '[Acetyl (K)]': 5,
    '[Acetyl (Protein N-term)]': 5,
    '[Carbamidomethyl (C)]': 3,
    '[Oxidation (M)]': 4,
    '[Phospho (STY)]': 0,
    '[GlyGly (K)]': 2,
    '[Methyl (KR)]': 6,
    '[Dimethyl (KR)]': 6,
    '[Trimethyl (K)]': 6,
    '[Pro5]': 9,
    '[Pro6]': 10,
    '[Glu->pyro-Glu]': 11,
    '[Gln->pyro-Glu]': 12,
    '[QQTGG (K)]': 13,
    '[Deamidation (N)]': 14,
    '[Deamidation 18O (N)]': 14,
    '[Deamidation (NQ)]': 14,
    '[Hydroxyproline]': 15,
    '[Carbamyl (N-term)]': 16,
    '[Delta:H(2)C(2) (N-term)]': 19,
    '[Dioxidation (MW)]': 4,
    '[Trioxidation (C)]': 4,
    '[Dethiomethyl (M)]': 20,
    '[Cation:Na (DE)]': 21,
    '[Methyl (E)]': 6,
    '[Dehydrated (ST)]': 23,
    '[Oxidation (P)]': 4,
    '[Dimethyl (K)]': 6,
    '[Amidated (Protein C-term)]': 22,
    '[Sulfo (STY)]': 24,
    '[Acetyl (N-term)]': 5,
    '[Amidated (C-term)]': 22,
    '[Sulfation (Y)]': 25,
    '[Phospho (ST)]': 0,
    '[Cys-Cys]': 26,
    '[Cysteinyl]': 27,
    '[Cysteinyl - carbamidomethyl]': 28,
    '[Oxidation (MP)]': 4
}

# Cell
import numpy as np
import pandas as pd
from pyteomics import fasta

def get_plot_data(protein,df,fasta):
    """
    Function to format experimental data for plotting.

    Args:
        protein (str): Uniprot protein accession.
        df (pd.DataFrame): Experimental data imported and formatted according to alphamap standards.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
    Returns:
        pd.DataFrame: Formatted dataframe for plotting.

    """
    protein_sequence = fasta[protein].sequence
    df_prot = df[df.unique_protein_id==protein]

    if df_prot.shape[0] == 0:
        df_plot = None
    else:
        df_peps = [np.arange(row['start'], row['end']+1) for _, row in df_prot.iterrows()]
        df_peps  = pd.DataFrame.from_records(data=df_peps)
        df_peps['modified_sequence'] = df_prot['modified_sequence'].values
        df_peps['all_protein_ids'] = df_prot['all_protein_ids'].values
        df_peps = df_peps.melt(id_vars=['modified_sequence','all_protein_ids'])
        df_peps = df_peps[['modified_sequence','all_protein_ids','value']].dropna()
        df_peps = df_peps.rename(columns={"value": "seq_position"})
        df_peps['marker_symbol'] = 1
        df_peps['marker_size'] = 8
        df_peps['PTM'] = np.NaN
        df_peps['PTMtype'] = np.NaN
        df_peps['PTMshape'] = np.NaN
        unique_pep = df_peps.modified_sequence.unique()
        for uid in unique_pep:
            df_peps_uid = df_peps[df_peps.modified_sequence==uid]
            start_uid = np.min(df_peps_uid.seq_position)
            end_uid = np.max(df_peps_uid.seq_position)
            df_peps['marker_symbol'] = np.where(df_peps.seq_position == start_uid, 7, df_peps.marker_symbol)
            df_peps['marker_symbol'] = np.where(df_peps.seq_position == end_uid, 8, df_peps.marker_symbol)
            df_peps['marker_size'] = np.where(df_peps.seq_position == start_uid, 6, df_peps.marker_size)
            df_peps['marker_size'] = np.where(df_peps.seq_position == end_uid, 6, df_peps.marker_size)

            df_PTMs_uid = df_prot[df_prot.modified_sequence==uid]
            #PTMsites = df_PTMs_uid.PTMsites.tolist()[0] + start_uid
            PTMsites = [x+start_uid for x in df_PTMs_uid.PTMsites.tolist()[0]]
            PTMtypes = df_PTMs_uid.PTMtypes.tolist()[0]

            #print(PTMsites)
            #print(PTMtypes)

            for i in range(0,len(PTMsites)):
                df_peps['PTM'] = np.where(df_peps["seq_position"]==PTMsites[i], 1, df_peps.PTM)
                df_peps['PTMtype'] = np.where(df_peps["seq_position"]==PTMsites[i], PTMtypes[i], df_peps.PTMtype)

            #print(df_peps)

            df_seq = pd.DataFrame({'seq_position':np.arange(0,len(protein_sequence))})

            df_plot = pd.merge(df_seq, df_peps, how='left', on='seq_position')
            df_plot['height']=0
            df_plot['color']="grey"

            unique_mods = df_plot['PTMtype'].dropna().unique()
            if len(unique_mods) > 0:
                for mod in df_plot['PTMtype'].dropna().unique():
                    if mod != 'nan':
                        #print(mod)
                        if mod not in ptm_shape_dict.keys():
                            ptm_shape_dict.update({mod : 17})

                        df_plot.loc[df_plot.PTMtype == mod, 'PTMshape'] = ptm_shape_dict[mod]
                        #df_plot.loc[df_plot.PTMtype == mod, 'PTMshape'] = 17

        #print(df_plot)

        df_plot['modified_sequence']= df_plot['modified_sequence'].astype(str)
        df_plot = df_plot.groupby('seq_position').agg({'modified_sequence':'; '.join,
                                                       'all_protein_ids': 'first',
                                                       'marker_symbol': 'first',
                                                       'marker_size': 'first',
                                                       'PTM': 'first',
                                                       'PTMtype': 'first',
                                                       'PTMshape': 'first',
                                                       'height': 'first',
                                                       'color': 'first'
                                                      }).reset_index()

        #print(df_plot)

    return(df_plot)

# Cell
import plotly.graph_objects as go

def plot_single_peptide_traces(df_plot,protein,fasta):
    """
    Function to plot single peptide trace.

    Args:
        df_plot (pd.DataFrame): Formatted dataframe for plotting, generated by get_plot_data.
        protein (str): Uniprot protein accession.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
    Returns:
        go.Figure: Figure data for a single dataset.

    """
    protein_sequence = fasta[protein].sequence
    entry_name = fasta[protein].description['GN']
    protein_name = fasta[protein].description['name']

    plot0 = go.Scatter(y=[None],
                       name='',
                       xaxis='x1',
                       showlegend=False)

    ## Peptide backbone
    df_plot_pep = df_plot.dropna(subset=['modified_sequence'])
    df_plot_pep = df_plot_pep[~df_plot_pep.modified_sequence.str.contains('nan')]
    plot1 = go.Scatter(x=df_plot_pep.seq_position+1,
                       y=df_plot.height,
                       xaxis='x2',
                       mode='markers',
                       marker_size=df_plot_pep.marker_size,
                       marker_symbol=df_plot_pep.marker_symbol,
                       marker_line_color=df_plot_pep.color,
                       marker_color=df_plot_pep.color,
                       marker_opacity=1,
                       meta=df_plot_pep.modified_sequence,
                       text=df_plot_pep.all_protein_ids,
                       hovertemplate ='Peptide: %{meta}<br>' +
                       'Protein IDs: %{text}',
                       name='',
                       showlegend=False)

    covered_AA = len(df_plot_pep.seq_position.unique())
    percent_AA_coverage = int(np.round(100/len(protein_sequence)*covered_AA))
    #print(percent_AA_coverage)

    ## PTM dots
    df_plot_ptm = df_plot.dropna(subset=['PTM'])
    #print(df_plot_ptm)
    plot2 = go.Scatter(x=df_plot_ptm.seq_position+1,
                       y=df_plot_ptm.height+0.3,
                       xaxis='x2',
                       mode='markers',
                       marker_size=8,
                       marker_symbol=df_plot_ptm.PTMshape,
                       marker_line_color=df_plot_ptm.color,
                       marker_color=df_plot_ptm.color,
                       marker_opacity=1,
                       text=df_plot_ptm.PTMtype,
                       hovertemplate = 'PTM: %{text}',
                       #hoverinfo='text',
                       name='',
                       showlegend=False)

    layout = go.Layout(
            yaxis=dict(
                title = "",
                ticks = None,
                showticklabels=False,
                range=[-1, 2],
                showgrid=False,
                zeroline=False
                ),
            xaxis1=dict(
                title= 'protein sequence',
                tickmode = 'array',
                range=[-10, len(protein_sequence)+10],
                tickvals = np.arange(1,len(protein_sequence)+1),
                ticktext = list(protein_sequence),
                tickangle=0,
                matches="x2",
                type="linear",
                anchor="y",
                showgrid=False,
                zeroline=False
                ),
            xaxis2=dict(
                title= 'AA position',
                tickmode = 'auto',
                range=[-10, len(protein_sequence)+10],
                tickangle=0,
                matches="x1",
                side="top",
                type="linear",
                anchor="y",
                showgrid=False,
                zeroline=False,
                tickformat = '.d'
                ),
        #showlegend=False,
        #height=400,
        #width=1000,
        plot_bgcolor='rgba(0,0,0,0)',
        title=f"Sequence plot for: {protein_name}<br>{entry_name} - {protein}",
        meta=percent_AA_coverage,
        margin = dict(l=20, r=20, t=150, b=20)
        )

    fig = go.Figure(data=[plot1,plot2,plot0], layout=layout)

    #print(fig.layout.meta)

    for i in range(0, df_plot_ptm.shape[0]):
            fig.add_shape(
                    dict(
                        type="line",
                        xref="x2",
                        x0=df_plot_ptm.seq_position.values[i] +1,
                        y0=df_plot_ptm.height.values[i],
                        x1=df_plot_ptm.seq_position.values[i] +1,
                        y1=df_plot_ptm.height.values[i]+0.3,
                        line=dict(
                            color=df_plot_ptm.color.values[i],
                            width=1
                        )
                    )
            )

    return fig

# Cell

# Color palettes used for the different uniprot annotation features
custom_color_palettes = {
    'col_greens':["#5C965D","#6AA16B","#77AC78","#84B786","#91C193","#9FCCA1","#B3DCB5","#C6EBC9"],
    'col_ornages':["#ff4800","#ff5400","#ff6000","#ff6d00","#ff7900","#ff8500","#ff9100","#ff9e00","#ffaa00","#ffb600"],
    'col_purples':["#ffa69e","#febaae","#fcb088","#d9f3e2","#b8f2e6","#aed9e0","#9baed9","#9199d5","#8783d1"],
    'col_turquises':["#00a9a5","#4e8098","#90c2e7"],
    'col_darkpinks':["#42033d","#6f0c59","#901468","#7c238c","#924ea6","#9c5eae"],
    'col_browns':["#5a2a27","#5c4742","#8d5b4c","#a5978b","#c4bbaf"]
}

# Cell

# The uniprot_color_dict maps each uniprot annotation feature to a color in the custom_color_palettes.
uniprot_color_dict = {'CHAIN': custom_color_palettes['col_greens'][0],
                      'INIT_MET': custom_color_palettes['col_greens'][1],
                      'PEPTIDE': custom_color_palettes['col_greens'][2],
                      'PROPEP': custom_color_palettes['col_greens'][3],
                      'SIGNAL': custom_color_palettes['col_greens'][4],
                      'TRANSIT': custom_color_palettes['col_greens'][5],

                      'COILED': custom_color_palettes['col_purples'][0],
                      'COMPBIAS': custom_color_palettes['col_purples'][1],
                      'DOMAIN': custom_color_palettes['col_purples'][2],
                      'MOTIF': custom_color_palettes['col_purples'][3],
                      'REGION': custom_color_palettes['col_purples'][4],
                      'REPEAT': custom_color_palettes['col_purples'][5],
                      'ZN_FING': custom_color_palettes['col_purples'][6],

                      'INTRAMEM': custom_color_palettes['col_turquises'][0],
                      'TOPO_DOM': custom_color_palettes['col_turquises'][1],
                      'TRANSMEM': custom_color_palettes['col_turquises'][2],

                      'STRUCTURE': 'black',
                      # extra structures
                      'Helix': '#5dabe8',
                      'Turn': '#e094bc',
                      'Beta strand': '#8cdbad',

                      'CROSSLNK': custom_color_palettes['col_ornages'][2],
                      'DISULFID': custom_color_palettes['col_ornages'][3],
                      'CARBOHYD': custom_color_palettes['col_ornages'][4],
                      'LIPID': custom_color_palettes['col_ornages'][5],
                      'MOD_RES': custom_color_palettes['col_ornages'][6],

                      'BINDING': custom_color_palettes['col_darkpinks'][0],
                      'CA_BIND': custom_color_palettes['col_darkpinks'][1],
                      'DNA_BIND': custom_color_palettes['col_darkpinks'][2],
                      'METAL': custom_color_palettes['col_darkpinks'][3],
                      'NP_BIND': custom_color_palettes['col_darkpinks'][4],
                      'SITE': custom_color_palettes['col_darkpinks'][5],

                      'NON_STD': custom_color_palettes['col_browns'][0],
                      'NON_CONS': custom_color_palettes['col_browns'][1],
                      'NON_TER': custom_color_palettes['col_browns'][2],
                      'VARIANT': custom_color_palettes['col_browns'][3],
                      'CONFLICT': custom_color_palettes['col_browns'][4],

                      'VAR_SEQ': '#fae7b1',
                      'UNSURE': 'grey',
                      'MUTAGEN': 'darkgrey',
                     }

# Cell

# Dictionary that maps one-letter amino acid abbreviations to their full name.
aa_color_dict = {'A':'Alanine',
                 'R':'Arginine',
                 'N':'Asparagine',
                 'D':'Aspartic acid',
                 'C':'Cysteine',
                 'E':'Glutamic acid',
                 'Q':'Glutamine',
                 'G':'Glycine',
                 'H':'Histidine',
                 'I':'Isoleucine',
                 'L':'Leucine',
                 'K':'Lysine',
                 'M':'Methionine',
                 'F':'Phenylalanine',
                 'P':'Proline',
                 'S':'Serine',
                 'T':'Threonine',
                 'W':'Tryptophan',
                 'Y':'Tyrosine',
                 'V':'Valine',
                 'X':'nan',
                 'U':'nan'}

# Cell
from structuremap.processing import download_alphafold_cif, download_alphafold_pae, format_alphafold_data, annotate_accessibility, get_smooth_score

# Cell
def get_quality_category(s):
    if s >= 90:
        n='Very high'
    elif s >= 70:
        n='Confident'
    elif s >= 50:
        n='Low'
    else:
        n='Very low'
    return(n)

# Cell
def get_exposure_category(s):
    if s > 10:
        n='Surrounded'
    elif s > 5:
        n='Many neighbors'
    elif s > 0:
        n='Few neighbors'
    else:
        n='No neighbors'
    return(n)

# Cell
def get_alphafold_annotation(protein: str,
                             selected_features: list,
                             download_folder: str = tempfile.gettempdir()) -> pd.DataFrame:

    alphafold_feature_dict = dict({'AlphaFold confidence':'quality',
        'AlphaFold exposure':'nAA_12_70_pae',
        'AlphaFold IDR':'AlphaFold IDR',
        'AlphaFold secondary structures':'structure_group'})

    alphafold_features = [i for i in [
        'AlphaFold confidence',
        'AlphaFold exposure',
        'AlphaFold IDR',
        'AlphaFold secondary structures'
    ] if i in selected_features]

    if len(alphafold_features) > 0:

        # download cif file
        valid_proteins, invalid_proteins, existing_proteins = download_alphafold_cif(
            proteins=[protein],
            out_folder=download_folder)
        is_available = np.sum([len(valid_proteins),len(existing_proteins)])

        if is_available > 0:

            af_data = format_alphafold_data(directory = download_folder, protein_ids = [protein])

            if len([i for i in ['AlphaFold exposure',
                                'AlphaFold IDR'] if i in selected_features]) > 0:
                # download pae file
                valid_pae, invalid_pae, existing_pae = download_alphafold_pae(
                    proteins=[protein],
                    out_folder=download_folder)
                is_available_pae = np.sum([len(valid_pae),len(existing_pae)])
                #@ToDo
                #if is_available == 0:
                #    print('error')

                # annotate exposure
                exposure_df = annotate_accessibility(df = af_data,
                                                     max_dist = 12,
                                                     max_angle = 70,
                                                     error_dir = download_folder)

                af_data = af_data.merge(exposure_df, how='left', on=['protein_id','AA','position'])

                # annotate exposure
                full_exposure_df = annotate_accessibility(df = af_data,
                                                          max_dist = 24,
                                                          max_angle = 180,
                                                          error_dir = download_folder)

                af_data = af_data.merge(full_exposure_df, how='left', on=['protein_id','AA','position'])

            # smooth scores
            smoothing_scores = []
            if len([i for i in ['AlphaFold IDR'] if i in alphafold_features]) > 0:
                smoothing_scores.append('nAA_24_180_pae')
            if len(smoothing_scores) > 0:
                af_data = get_smooth_score(af_data, np.array(smoothing_scores),[10])

            # annotate predicted IDRs
            if 'AlphaFold IDR' in alphafold_features:
                af_data['AlphaFold IDR'] = np.where(af_data['nAA_24_180_pae_smooth10']<=34.27, "IDR", "folded")

            af_columns = ['position','AA']
            af_columns = af_columns + [alphafold_feature_dict[i] for i in alphafold_features]
            af_data = af_data[af_columns]

            if 'quality' in af_data.columns:
                af_data["AlphaFold confidence"] = [get_quality_category(x) for x in af_data.quality]

            if 'nAA_12_70_pae' in af_data.columns:
                af_data["AlphaFold exposure"] = [get_exposure_category(x) for x in af_data.nAA_12_70_pae]

            if 'structure_group' in af_data.columns:
                af_data = af_data.rename(columns={"structure_group": "AlphaFold secondary structures"})

        else:
            af_data = pd.DataFrame({'AA' : []})
    else:
        af_data = pd.DataFrame({'AA' : []})

    return(af_data)

# Cell

import plotly.graph_objects as go
from .proteolytic_cleavage import get_cleavage_sites

def plot_peptide_traces(df: pd.DataFrame or list,
                        name: str or list,
                        protein: str,
                        fasta: fasta,
                        uniprot: pd.DataFrame,
                        selected_features: list,
                        uniprot_feature_dict: dict,
                        uniprot_color_dict: dict,
                        selected_proteases: list = [],
                        selected_alphafold_features: list = [],
                        dashboard: bool = False,
                        trace_colors: list = [],
                        download_folder: str = tempfile.gettempdir()):

    """
    Function to generate the sequence plot.

    Args:
        df (pd.DataFrame/list): Single dataframe or list of dataframes containing the datasets to plot.
        name (str/list): Single string or list of strings containing the names for each dataset in df.
        protein (str): Uniprot protein accession.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
        uniprot (pd.DataFrame): Uniprot annotations formatted by alphamap.
        selected_features (list): List of uniprot features to plot.
        uniprot_feature_dict (dict): Uniprot feature dictionary.
        uniprot_color_dict (dict): Uniprot color dictionary.
        selected_proteases (list, optional): List of proteases to plot. Default is an empty list.
        dashboard (bool, optional): Flag if the function is called from the dashboard. Default is 'False'.
        trace_colors (list, optional): List of manualy selected colors for each dataset in df. Default is an empty list.

    Returns:
        go.Figure: Sequence plot.

    """

    figure_height = 200

    protein_sequence = fasta[protein].sequence

    # colors for experimental data traces
    if len(trace_colors) == 0:
        colors = ["#023e8a","#0096c7","#90e0ef","#7fd14d","#26a96c",
                  "#0D7F8B", "#24B4AB", "#9DE7BE", "#1B9CC6", "#016699"]
    else:
        colors = trace_colors

    # generation of a reverse uniprot_feature_dict
    uniprot_feature_dict_rev = {v: k for k, v in uniprot_feature_dict.items()}
    #uniprot_feature_dict_rev["STRUCTURE"] = "Secondary structure"

    # subsetting of the uniprot annotation to the selected features
    uniprot_annotation_p = uniprot[uniprot.protein_id==protein]
    # formatting of uniprot annotations
    uniprot_annotation_p_f = format_uniprot_annotation(uniprot_annotation_p, uniprot_feature_dict)
    # subset for selected features
    uniprot_annotation_p_f_f = uniprot_annotation_p_f[uniprot_annotation_p_f.feature.isin(selected_features)]

    #@ToDo if

    if isinstance(df, pd.DataFrame):
        df_plot = get_plot_data(protein=protein,
                              df = df,
                              fasta = fasta)

        df_plot.color = colors[0]

        observed_mods = list(set(df_plot.PTMtype))
        ptm_shape_dict_sub = {key: ptm_shape_dict[key] for key in observed_mods if key in ptm_shape_dict}

        fig = plot_single_peptide_traces(df_plot,protein=protein,fasta = fasta)

        AA_coverage = fig.layout.meta
        trace_name = [name + "<br> (" + str(AA_coverage) + "% coverage)"]

        fig.update_layout(yaxis=dict(showticklabels=True,
                                     tickmode = 'array',
                                     tickvals = [0],
                                     ticktext = [name + "(" + str(AA_coverage) + "%)"],
                                     showgrid=False))

        y_max = 1

    elif isinstance(df, list):

        df_plot = [get_plot_data(protein=protein,
                               df = d,
                               fasta = fasta) for d in df]

        # Subset data and annotations for the samples where the selected protein was detected
        valid_idx = []
        for i in range(len(df_plot)):
            if df_plot[i] is not None:
                valid_idx.append(i)
        df_plot = [df_plot[i] for i in valid_idx]
        name = [name[i] for i in valid_idx]
        colors = [colors[i] for i in valid_idx]
        #observed_mods = set([df_plot[i].PTMtype for i in valid_idx])
        observed_mods = []
        for i in range(len(df_plot)):
            observed_mods.extend(list(set(df_plot[i].PTMtype)))
        observed_mods = list(set(observed_mods))
        ptm_shape_dict_sub = {key: ptm_shape_dict[key] for key in observed_mods if key in ptm_shape_dict}

        for i in range(len(df_plot)):
            df_plot[i].color = colors[i]
            df_plot[i].height = 1+i

        plot_list = [plot_single_peptide_traces(df,protein=protein,fasta = fasta) for df in df_plot]
        new_data = [p.data for p in plot_list]
        new_data = sum(new_data, ())
        new_layout = plot_list[0].layout
        shapes = [p.layout.shapes for p in plot_list]
        shapes = sum(shapes, ())
        new_layout.shapes = new_layout.shapes + tuple(shapes)
        AA_coverage = [p.layout.meta for p in plot_list]
        trace_name = [n + "<br> (" + str(c) + "% coverage)" for n,c in zip(name,AA_coverage)]

        fig = go.Figure(data=new_data, layout=new_layout)
        fig.update_layout(yaxis=dict(range=[0,len(df_plot)+1],
                                     showticklabels=True,
                                     tickmode = 'array',
                                     tickvals = np.arange(0, len(df_plot))+1,
                                     ticktext = np.array(trace_name),
                                     showgrid=False))

        y_max = len(df_plot)+1

        figure_height = figure_height + (len(df_plot)*50)


    ptm_shape_dict_sub = dict(sorted(ptm_shape_dict_sub.items()))
    for i in range(len(ptm_shape_dict_sub)):
        fig.add_trace(go.Scatter(y=[None],
                                 mode='markers',
                                 xaxis='x2',
                                 marker=dict(symbol=list(ptm_shape_dict_sub.values())[i],
                                             color='black'),
                                 name=list(ptm_shape_dict_sub.keys())[i],
                                 showlegend=True))

    all_uniprot_features = list(uniprot_color_dict.keys())
    available_features = list(set(uniprot_annotation_p_f_f.feature))
    unique_features = [x for x in all_uniprot_features if x in available_features]
    if len(unique_features) > 0:

        y_max = y_max+1

        for j in range(0,len(unique_features)):

            figure_height = figure_height + 50

            domain = unique_features[j]
            domain_info_sub = uniprot_annotation_p_f_f[uniprot_annotation_p_f_f.feature==domain].reset_index(drop=True)
            for i in range(0, domain_info_sub.shape[0]):
                start = int(domain_info_sub.start[i])
                end = domain_info_sub.end[i]
                if np.isnan(domain_info_sub.end[i]):
                    end=start #+1
                else:
                    end=int(end)

                if domain_info_sub.feature[i] == "STRUCTURE":
                    marker_col = uniprot_color_dict[domain_info_sub.annotation[i]]
                else:
                    marker_col = uniprot_color_dict[domain_info_sub.feature[i]]

                fig.add_trace(go.Bar(x=list(range(start,end+1)),
                                     y=list(np.repeat(0.2,end-start+1)),
                                     base=list(np.repeat(y_max+(j/2),end-start+1)-0.1),
                                     marker_color=marker_col,
                                     marker_line_width=0,
                                     opacity=0.8,
                                     showlegend=False,
                                     xaxis='x2',
                                     name='',
                                     text=np.repeat(domain_info_sub.annotation[i],len(range(start,end+1))),
                                     hovertemplate ='%{text}'
                                     #hovertext=domain_info_sub.annotation[i],
                                     #hoverinfo='text'
                                    ))
        fig.update_layout(barmode='stack', bargap=0, hovermode='x unified',hoverdistance=1)

        y_max = y_max + (len(unique_features)/2)

    selected_proteases = sorted(selected_proteases)
    if len(selected_proteases) > 0:

        y_max = y_max+1

        for u in range(0,len(selected_proteases)):

            figure_height = figure_height + 50

            protease = selected_proteases[u]
            sites = get_cleavage_sites(protein_sequence,protease)
            for s in sites:
                fig.add_trace(go.Bar(x=list(range(s+1,s+2)),
                                     y=[0.2],
                                     base=y_max+(u/2)-0.1,
                                     marker_color="grey",
                                     opacity=0.8,
                                     showlegend=False,
                                     xaxis='x2',
                                     name='',
                                     text=np.repeat(protease,len(range(s+1,s+2))),
                                     hovertemplate ='%{text}'
                                     #hovertext=protease,
                                     #hoverinfo='text'
                                    ))

        y_max = y_max + (len(selected_proteases)/2)


    alphafold_annotation = get_alphafold_annotation(protein = protein,
                                                    selected_features = selected_alphafold_features,
                                                    download_folder = download_folder)

    if alphafold_annotation.empty:
        #print('no alphafold')
        alphafold_features = []
    else:
        #print('alphafold')

        alphafold_features = [i for i in alphafold_annotation.columns if i in [
            "AlphaFold confidence",
            "AlphaFold exposure",
            "AlphaFold IDR",
            "AlphaFold secondary structures"]]

        alphafold_color_dict = {'HELX': '#5dabe8',
                                'STRN': '#8cdbad',
                                'TURN': '#e094bc',
                                'BEND': '#fac682',
                                'unstructured': '#aaaaaa',
                                'Very high':'#4035de',
                                'Confident':'#6aaff7',
                                'Low':'#f5d44e',
                                'Very low':'#ff8f4a',
                                'No neighbors': "#8FF5A2",
                                'Few neighbors': "#38BFA7",
                                'Many neighbors': "#367BC3",
                                'Surrounded': "#524582",
                                'IDR': "#B5B5B5",
                                'folded': "#367BC3"}

        y_max = y_max+1

        for j in range(0,len(alphafold_features)):

            figure_height = figure_height + 50

            af_column = alphafold_features[j]
            if af_column == "AlphaFold confidence":
                extra_annotation = alphafold_annotation["quality"].values
            elif af_column == "AlphaFold exposure":
                extra_annotation = alphafold_annotation["nAA_12_70_pae"].values
            else:
                extra_annotation = np.repeat("",alphafold_annotation.shape[0])

            for i in range(0, alphafold_annotation.shape[0]):
                start = alphafold_annotation.position[i]
                end = alphafold_annotation.position[i]

                alpha_col = alphafold_color_dict[alphafold_annotation[af_column][i]]

                fig.add_trace(go.Bar(x=list(range(start,end+1)),
                                     y=list(np.repeat(0.2,end-start+1)),
                                     base=y_max+(j/2)-0.1,
                                     marker_color=alpha_col,
                                     marker_line_width=0,
                                     opacity=0.8,
                                     showlegend=False,
                                     xaxis='x2',
                                     name='',
                                     text=np.repeat(str(alphafold_annotation[af_column][i]) + " " + str(extra_annotation[i]),len(range(start,end+1))),
                                     hovertemplate ='%{text}'
                                     #hovertext=domain_info_sub.annotation[i],
                                     #hoverinfo='text'
                                    ))

        y_max = y_max + (len(alphafold_features)/2)

    fig.add_trace(go.Scatter(x=np.arange(1,len(protein_sequence)+1,1),
                        y=np.repeat(0,len(protein_sequence)),
                        marker=dict(color='rgba(135, 206, 250, 0)'),
                        name='',
                        mode='markers',
                        xaxis='x2',
                        text=[aa_color_dict[x] for x in list(protein_sequence)],
                        #text=np.arange(1,len(protein_sequence)+1,1),
                        meta=list(protein_sequence),
                        hovertemplate ='<b>%{meta}: %{text}<b>',
                        showlegend=False))

    if figure_height < 500:
        figure_height = 500

    if dashboard:
        plot_width = None
    else:
        plot_width = 1500

    fig.update_layout(barmode='stack', bargap=0, hovermode='x unified',hoverdistance=1,
                      width=plot_width, height=figure_height)

    mapped_feature_names = [uniprot_feature_dict_rev.get(key) for key in unique_features]
    if isinstance(df, pd.DataFrame):
        trace_ax = list([0])
        trace_count = 1
    elif isinstance(df, list):
        trace_ax = list(np.arange(1,len(df_plot)+1,1))
        trace_count = len(df_plot) + 1
    #print(trace_ax)

    if len(unique_features) > 0:
        trace_count += 1
        uniprot_ax = list(np.arange(0,len(unique_features)/2, 0.5) + trace_count)
        trace_count = max(uniprot_ax)+0.5
        #print(uniprot_ax)
    else:
        uniprot_ax = []
    if len(selected_proteases) > 0:
        trace_count += 1
        protease_ax = list(np.arange(0, len(selected_proteases)/2,0.5) + trace_count)
        trace_count = max(protease_ax)+0.5
        #print(protease_ax)
    else:
        protease_ax = []
    if len(alphafold_features) > 0:
        trace_count += 1
        alphafold_ax = list(np.arange(0, len(alphafold_features)/2,0.5) + trace_count)
        trace_count = max(alphafold_ax)+0.5
        #print(alphafold_ax)
    else:
        alphafold_ax = []

    fig.update_yaxes(showticklabels=True,
                     #tickvals= np.arange(0, 1+len(unique_features)+len(selected_proteases)),
                     tickvals= trace_ax + uniprot_ax + protease_ax + alphafold_ax,
                     ticktext=np.hstack((np.array(trace_name),np.array(mapped_feature_names),np.array(selected_proteases),np.array(alphafold_features))),
                     automargin=True,
                     range=[-1, y_max + 0.2],
                     showgrid=False)

    #config = {'toImageButtonOptions': {'format': 'svg', # one of png, svg, jpeg, webp
    #                                   'filename': 'custom_image',
    #                                   'height': 500,
    #                                   'width': 1500,
    #                                   'scale': 1 # Multiply title/legend/axis/canvas sizes by this factor
    #                                  }
    #         }

    return fig #.show(config=config)

# Cell
def extract_annotation(df: pd.DataFrame):
    df = df[df.marker_symbol>0]
    df = df.astype({"PTMtype": str})
    df['MS data'] = ['detected' if m=='nan' else m for m in df.PTMtype]
    df = df[['seq_position','MS data']]
    df['MS data'] = [x.replace(" ", "") for x in df['MS data']]
    df = df.rename(columns={"seq_position": "position"})
    df['position'] = df['position']+1
    return df

# Cell
def manipulate_cif(protein: str,
                   MS_data: pd.DataFrame,
                   download_folder: str = tempfile.gettempdir(),
                   ):


    print(MS_data)

    alphafold_annotation = get_alphafold_annotation(protein = protein,
                                                    selected_features = ["AlphaFold confidence",
                                                                         "AlphaFold exposure",
                                                                         "AlphaFold IDR",
                                                                         "AlphaFold secondary structures"],
                                                    download_folder = download_folder)

    if alphafold_annotation.empty:
        return False

    else:
        mod_table = False
        AA_position = 0
        infile = os.path.join(download_folder, protein+'.cif')
        outfile = os.path.join(download_folder, protein+'_mod.cif')

        with open(infile) as f:
            lines = []
            for line in f:
                line = line.rstrip()
                if mod_table:
                    if line == '#':
                        mod_table = False
                        lines.append(line)
                    else:
                        AA_position += 1

                        # append exposure
                        exposure = str(alphafold_annotation[alphafold_annotation.position==AA_position].nAA_12_70_pae.values[0])
                        n_extra_blank = 4-len(exposure)
                        appended_line = line+' '+exposure+(n_extra_blank*' ')
                        # append MS data
                        if AA_position in list(MS_data.position.values):
                            appended_line = appended_line+' '+'detected'+' '
                            ms_evidence = MS_data[MS_data.position==AA_position]['MS data'].values[0]
                            n_extra_blank = 40-len(ms_evidence)
                            appended_line = appended_line+' '+ms_evidence+(n_extra_blank*' ')
                        else:
                            appended_line = appended_line+' '+'absent  '+' '
                            n_extra_blank = 40-len('absent')
                            appended_line = appended_line+' '+'absent'+(n_extra_blank*' ')
                        # append PTM data
                        ptm_sub = MS_data[MS_data['MS data'] != 'detected']
                        if AA_position in list(ptm_sub.position.values):
                            ptm_evidence = ptm_sub[ptm_sub.position==AA_position]['MS data'].values[0]
                            n_extra_blank = 40-len(ptm_evidence)
                            appended_line = appended_line+' '+ptm_evidence+(n_extra_blank*' ')
                        else:
                            n_extra_blank = 40-len('none')
                            appended_line = appended_line+' '+'none'+(n_extra_blank*' ')
                        # append IDR
                        IDR_info = alphafold_annotation[alphafold_annotation.position==AA_position]['AlphaFold IDR'].values[0]
                        n_extra_blank = 10-len(IDR_info)
                        appended_line = appended_line+' '+IDR_info+(n_extra_blank*' ')
                        # append IDR + PTM data
                        if AA_position in list(ptm_sub.position.values):
                            n_extra_blank = 40-len(ptm_evidence)
                            appended_line = appended_line+' '+ptm_evidence+(n_extra_blank*' ')
                        else:
                            n_extra_blank = 40-len(IDR_info)
                            appended_line = appended_line+' '+IDR_info+(n_extra_blank*' ')
                        # append structure group
                        structure_info = alphafold_annotation[alphafold_annotation.position==AA_position]['AlphaFold secondary structures'].values[0]
                        n_extra_blank = 15-len(structure_info)
                        appended_line = appended_line+' '+structure_info+(n_extra_blank*' ')

                        lines.append(appended_line)
                else:
                    if line == '_ma_qa_metric_local.ordinal_id':
                        lines.append(line)
                        lines.append('_ma_qa_metric_local.exposure')
                        lines.append('_ma_qa_metric_local.aa')
                        lines.append('_ma_qa_metric_local.aa_ptm')
                        lines.append('_ma_qa_metric_local.ptm')
                        lines.append('_ma_qa_metric_local.idr')
                        lines.append('_ma_qa_metric_local.idr_ptm')
                        lines.append('_ma_qa_metric_local.structure_group')
                        mod_table = True
                    else:
                        lines.append(line)

        try:
            os.remove(outfile)
        except:
            pass

        with open(outfile, 'a') as outfile:
            for l in lines:
                outfile.write(l+'\n')

        return True

# Cell
def adjust_html(protein: str,
                coloring: str,
                data_name: str):

    html_template = """
                    <!DOCTYPE html>
                    <html lang="en">
                      <head>
                        <meta charset="utf-8" />
                        <meta name="viewport" content="width=device-width, user-scalable=no, minimum-scale=1.0, maximum-scale=1.0">
                        <title>PDBe Molstar</title>

                        <!-- Molstar CSS & JS -->
                        <link rel="stylesheet" type="text/css" href="js/alphafold-viz.css">
                        <script type="text/javascript" src="js/alphafold-viz.js"></script>
                        <style>
                          * {
                              margin: 0;
                              padding: 0;
                              box-sizing: border-box;
                          }
                          .msp-plugin ::-webkit-scrollbar-thumb {
                              background-color: #5b5b68 !important;
                          }
                          .viewerSection {
                            padding-top: 100px;
                          }
                          .controlsSection {
                            width: 300px;
                            float: left;
                            padding: 40px 0 0 40px;
                            margin-right: 30px;
                          }
                          .controlBox {
                            border: 1px solid lightgray;
                            padding: 10px;
                            margin-bottom: 20px;
                          }
                          #myViewer{
                            float:left;
                            width: 800px;
                            height: 800px;
                            position:relative;
                          }
                        </style>
                      </head>

                      <body>



                        <div class="viewerSection">
                          <!-- <h4>3D viewer</h4> -->
                          <h4>   </h4>

                          <!-- Molstar container -->
                          <div id="myViewer"></div>

                        </div>
                        <script>

                          globalThis.color_choice = your_choice //'af', 'access', 'ptm'

                          //Create plugin instance
                          var viewerInstance = new PDBeMolstarPlugin();

                          //Set options (Checkout available options list in the documentation)
                          var options = {
                            customData: {
                              url: 'cif/your_file.cif',
                              format: 'cif'
                            },
                            hideControls: false,
                            // expanded: true,
                            isAfView: true,
                            hideCanvasControls: [] //['selection', 'animation', 'controlToggle', 'controlInfo']
                          }

                          //Get element from HTML/Template to place the viewer
                          var viewerContainer = document.getElementById('myViewer');

                          //Call render method to display the 3D view
                          viewerInstance.render(viewerContainer, options);
                          viewerInstance.canvas.setBgColor({r:255, g:255, b:255});

                        </script>
                      </body>

                    </html>
                    """

    html_template = html_template.replace('your_file.cif',protein+'_'+data_name+'_mod.cif')

    if coloring == 'AlphaFold confidence':
        color = 'af'
    elif coloring == 'AlphaFold exposure':
        color = 'exposure'
    elif coloring == 'MS peptides':
        color = 'aa'
    elif coloring == 'MS modified peptides':
        color = 'aa_ptm'
    elif coloring == 'MS PTMs':
        color = 'ptm'
    elif coloring == 'AlphaFold IDR':
        color = 'idr'
    elif coloring == 'MS PTMs on AlphaFold IDR':
        color = 'idr_ptm'
    elif coloring == 'AlphaFold secondary structures':
        color = 'structure_group'
    else:
        color = 'af'

    html = html_template.replace('globalThis.color_choice = your_choice',f"globalThis.color_choice = '{color}'")

    return html

# Cell
def get_ms_concensus(ms_list):
    unique_ms_list = set(ms_list)
    if len(unique_ms_list)==1:
        out = unique_ms_list
    else:
        out = [i for i in unique_ms_list if i != 'detected']

    out = ';'.join(out)

    return out

# Cell
import alphamap
def plot_3d_structure(df: pd.DataFrame or list,
                        name: str,
                        protein: str,
                        fasta: fasta,
                        selected_coloring: str,
                        dashboard: bool = False,
                        download_folder: str = tempfile.gettempdir()):

    """
    Function to generate the sequence plot.

    Args:
        df (pd.DataFrame/list): Single dataframe or list of dataframes containing the datasets to plot.
        name (str): Single string containing the name of the MS dataset in df.
        protein (str): Uniprot protein accession.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
        selected_coloring (str): Coloring to show.
        dashboard (bool, optional): Flag if the function is called from the dashboard. Default is 'False'.

    Returns:
        go.Figure: 3D plot.

    """

    if isinstance(df, pd.DataFrame):
        df_plot = get_plot_data(protein=protein,
                                df = df,
                                fasta = fasta)
    elif isinstance(df, list):
        df_plot = [get_plot_data(protein=protein,
                                 df = d,
                                 fasta = fasta) for d in df]
        df_plot = pd.concat(df_plot)

    if df_plot is None:
        MS_annotation =  pd.DataFrame({'position' : [], 'MS data': []})
    else:
        MS_annotation = extract_annotation(df_plot)
        MS_annotation = MS_annotation.groupby('position')['MS data'].apply(list).reset_index(name='MS data')
        MS_annotation['MS data'] = MS_annotation['MS data'].apply(lambda x: get_ms_concensus(x))

    cif_available = manipulate_cif(protein = protein,
                                   MS_data = MS_annotation,
                                   download_folder = download_folder)

    BASE_PATH = os.path.dirname(alphamap.__file__)
    js_path = os.path.join(BASE_PATH, 'js')
    cif_path = download_folder


    import shutil
    new_file = os.path.join(cif_path, protein+'_'+name+'_mod.cif')
    try:
        os.remove(new_file)
    except:
        pass
    shutil.copyfile(src = os.path.join(cif_path, protein+'_mod.cif'),
                    dst = new_file)

    if cif_available:
        mod_html = adjust_html(protein=protein,
                               coloring=selected_coloring,
                               data_name=name)
    else:
        mod_html = f"""
            <!DOCTYPE html>
            <center> No AlphaFold structure available for {protein}. <center>
            """



    return mod_html, js_path, cif_path

# Cell
from .pdflib import *

def create_pdf_report(proteins: list,
                      df: pd.DataFrame or list,
                      name: str or list,
                      fasta: fasta,
                      uniprot: pd.DataFrame,
                      selected_features: list,
                      uniprot_feature_dict: dict,
                      uniprot_color_dict: dict,
                      selected_proteases: list = [],
                      trace_colors: list = []):
    """
    Function to write pdf reports for selected proteins

    Args:
        proteins (list): List of uniprot protein accessions.
        df (pd.DataFrame/list): Single dataframe or list of dataframes containing the datasets to plot.
        name (str/list): Single string or list of strings containing the names for each dataset in df.
        fasta (fasta): Fasta file imported by pyteomics 'fasta.IndexedUniProt'.
        uniprot (pd.DataFrame): Uniprot annotations formatted by alphamap.
        selected_features (list): List of uniprot features to plot.
        uniprot_feature_dict (dict): Uniprot feature dictionary.
        uniprot_color_dict (dict): Uniprot color dictionary.
        selected_proteases (list, optional): List of proteases to plot. Default is an empty list.
        trace_colors (list, optional): List of manualy selected colors for each dataset in df. Default is an empty list.

    Returns:
        BytesIO: BytesIO object for writing a pdf report.
    """

    if isinstance(df, pd.DataFrame):
        max_height = 200 + 50 + (len(selected_features)*50) + (len(selected_proteases)*50)
    else:
        max_height = 200 + (len(df)*50) + (len(selected_features)*50) + (len(selected_proteases)*50)


    if max_height < 700:
        max_height = 700

    footer_text = '<font size="20">This report was generated by <a href="https://github.com/MannLabs/alphamap" color="darkblue"><b>AlphaMap</b></a>.</font>'

    pdf_buf = BytesIO()
    pdf_report = canvas.Canvas(pdf_buf, pagesize=(1600,max_height))

    if isinstance(proteins, list):
        for p in range(0,len(proteins)):
            plot = plot_peptide_traces(df=df, name=name, protein=proteins[p], fasta=fasta,
                                       uniprot=uniprot, selected_features=selected_features,
                                       uniprot_feature_dict=uniprot_feature_dict,
                                       uniprot_color_dict=uniprot_color_dict,
                                       selected_proteases=selected_proteases,
                                       trace_colors=trace_colors)
            draw_content(pdf_report, plot, width=1600, height=max_height,
                         spacing=5, border=20)
            draw_content(pdf_report, footer_text, width=1600, height=100,
                         spacing=20, border=30)
            pdf_report.showPage()
    else:
        plot = plot_peptide_traces(df=df, name=name, protein=proteins, fasta=fasta,
                                   uniprot=uniprot, selected_features=selected_features,
                                   uniprot_feature_dict=uniprot_feature_dict,
                                   uniprot_color_dict=uniprot_color_dict,
                                   selected_proteases=selected_proteases,
                                   trace_colors=trace_colors)
        draw_content(pdf_report, plot, width=1600, height=max_height,
                         spacing=20, border=30)
        draw_content(pdf_report, footer_text, width=1600, height=100,
                         spacing=20, border=30)
        pdf_report.showPage()

    pdf_report.save()
    pdf_buf.seek(0)

    #with open("file.pdf", "wb") as file:
    #    file.write(pdf_buf.getvalue())

    return pdf_buf


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/Uniprot_integration.ipynb (unless otherwise specified).

__all__ = ['extract_note', 'extract_note_end', 'resolve_unclear_position', 'extract_positions', 'preprocess_uniprot',
           'uniprot_feature_dict']

# Cell
import re

import pandas as pd
import numpy as np

# Cell
def extract_note(string: str, splitted:bool = False):
    """
    Helper function to extract information about note of the protein from Uniprot using regular expression.

    Args:
        string (str): Uniprot annotation string.
        splitted (bool, optional): Flag to allow linebreaks. Default is 'False'.
    Returns:
        str: Extracted string of the uniprot note section.
    """
    if not splitted:
        regex = r"\/note=\"(?P<note>.+?)\""
    else:
        regex = r"\/note=\"(?P<note>.*)"
    result = re.findall(regex, string)
    return result

def extract_note_end(string: str, has_mark:bool = True):
    """
    Helper function to extract information about note of the protein from Uniprot using regular expression.

    Args:
        string (str): Uniprot annotation string.
        has_mark (bool, optional): Flag if end quotation marks are present. Default is 'False'.
    Returns:
        str: Extracted string of the uniprot note section.
    """
    if has_mark:
        regex = r"FT\s+(?P<note>.*)\""
    else:
        regex = r"FT\s+(?P<note>.*)"
    result = re.findall(regex, string)
    return result

# Cell
def resolve_unclear_position(value: str):
    """
    Replace unclear position of the start/end of the modification defined as '?' with -1 and if it's defined as '?N'
    or ">N" - by removing the '?'/'>'/'<' signs.

    Args:
        value (str): Unclear sequence position from uniprot.
    Returns:
        float: Resolved sequence position.
    """
    # if it's "1..?" or "?..345" for start or end -> remove -1 that we can filter later
    # if it's "31..?327" or "?31..327" -> remove the question mark
    # if it's "<1..106" or "22..>115" -> remove the "<" or ">" signs
    if value == '?':
        return -1
    value = value.replace('?', '').replace('>', '').replace('<', '')
    return float(value)

def extract_positions(posit_string: str):
    """
    Extract isoform_id(str) and start/end positions(float) of any feature key from the string.

    Args:
        posit_string (str): Uniprot position string.
    Returns:
        [str, float, float]: str: Uniprot isoform accession, float: start position, float: end position
    """
    isoform = ''
    start = end = np.nan
    if '..' in posit_string:
        start, end = posit_string.split('..')
    if ':' in posit_string:
        if isinstance(start, str):
            isoform, start = start.split(':')
        else:
            isoform, start = posit_string.split(':')
    # in the case when we have only one numeric value as a posit_string
    if isinstance(start, float):
        start = posit_string
    # change the type of start and end into int/float(np.nan)
    if isinstance(start, str):
        start = resolve_unclear_position(start)
    if isinstance(end, str):
        end = resolve_unclear_position(end)
    return isoform, start, end

# Cell
def preprocess_uniprot(path_to_file: str):
    """
    A complex complete function to preprocess Uniprot data from specifying the path to a flat text file
    to the returning a dataframe containing information about:
        - protein_id(str)
        - feature(category)
        - isoform_id(str)
        - start(float)
        - end(float)
        - note information(str)

    Args:
        path_to_file (str): Path to a .txt annotation file directly downloaded from uniprot.
    Returns:
        pd.DataFrame: Dataframe with formatted uniprot annotations for alphamap.

    """
    all_data = []
    with open(path_to_file) as f:

        is_splitted = False
        new_instance = False
        combined_note = []
        line_type = ''

        for line in f:

            if line.startswith(('AC', 'FT')):
                if is_splitted:
                    # in case when the note information is splitted into several lines
                    if line.rstrip().endswith('"'):
                        # if it's the final part of the note
                        combined_note.extend(extract_note_end(line))
                        all_data.append([protein_id, feature, isoform, start, end, " ".join(combined_note)])
                        is_splitted = False
                        new_instance = False
                    else:
                        # if it's the middle part of the note
                        combined_note.extend(extract_note_end(line, has_mark=False))
                elif line.startswith('AC'):
                    # contains the protein_id information
                    if line_type != 'AC':
                        # to prevent a situation when the protein has several AC lines with different names
                        # in this case we are taking the first name in the first line
                        protein_id = line.split()[1].replace(';', '')
                    line_type = 'AC'
                elif line.startswith('FT'):
                    line_type = 'FT'
                    # contains all modifications/preprocessing events/etc., their positions, notes
                    data = line.split()
                    if data[1].isupper() and not data[1].startswith('ECO'):
                            feature = data[1]
                            isoform, start, end = extract_positions(data[2])
                            new_instance = True
                    else:
                        if data[1].startswith('/note'):
                            note = extract_note(line)
                            if note:
                                # if note was created > it contains just one line and can be already added to the data
                                all_data.append([protein_id, feature, isoform, start, end, note[0]])
                                new_instance = False
                            else:
                                # if note is empty > it's splitted into several lines and we create combined_note
                                combined_note = extract_note(line, splitted=True)
                                is_splitted = True
                        else:
                            if new_instance:
                                # in case when we don't have any note but need to add other information about instance
                                all_data.append([protein_id, feature, isoform, start, end, ''])
                                new_instance = False

    # create a dataframe for preprocessed data
    uniprot_df = pd.DataFrame(all_data, columns=['protein_id', 'feature', 'isoform_id', 'start', 'end', 'note'])
    # change the dtypes of the columns
    uniprot_df.feature = uniprot_df.feature.astype('category')
    # to filter the instances that don't have a defined start/end position(start=-1 or end=-1)
    uniprot_df = uniprot_df[(uniprot_df.start != -1) & (uniprot_df.end != -1)].reset_index(drop=True)

    return uniprot_df

# Cell
uniprot_feature_dict = {
    'Chain': 'CHAIN',
    'Initiator methionine': 'INIT_MET',
    'Peptide': 'PEPTIDE',
    'Propeptide': 'PROPEP',
    'Signal peptide': 'SIGNAL',
    'Transit peptide': 'TRANSIT',
    'Cross-link': 'CROSSLNK',
    'Disulfide bond': 'DISULFID',
    'Glycosylation': 'CARBOHYD',
    'Lipidation': 'LIPID',
    'Modified residue': 'MOD_RES',
    'Coiled coil': 'COILED',
    'Compositional bias': 'COMPBIAS',
    'Domain': 'DOMAIN',
    'Motif': 'MOTIF',
    'Region': 'REGION',
    'Repeat': 'REPEAT',
    'Zinc finger': 'ZN_FING',
    'Intramembrane': 'INTRAMEM',
    'Topological domain': 'TOPO_DOM',
    'Transmembrane': 'TRANSMEM',
    'Beta strand': 'STRAND',
    'Helix': 'HELIX',
    'Turn': 'TURN',
    'Active site': 'ACT_SITE',
    'Binding site': 'BINDING',
    'Calcium binding': 'CA_BIND',
    'DNA binding': 'DNA_BIND',
    'Metal binding': 'METAL',
    'Nucleotide binding': 'NP_BIND',
    'Site': 'SITE',
    'Non-standard residue': 'NON_STD',
    'Non-adjacent residues': 'NON_CONS',
    'Non-terminal residue': 'NON_TER',
    'Natural variant': 'VARIANT',
    'Sequence conflict': 'CONFLICT',
    'Alternative sequence': 'VAR_SEQ',
    'Sequence uncertainty': 'UNSURE',
    'Secondary structure': 'STRUCTURE',
    'Mutagenesis': 'MUTAGEN'
}


#!python

# external
import os
import platform
import urllib.request
import shutil
import re
import sys
import numpy as np
import pandas as pd
from io import StringIO
import pyteomics.fasta
# visualization libraries
import panel as pn
pn.extension()
import plotly.graph_objects as go
# local
from alphamap.importing import import_data, extract_rawfile_unique_values
from alphamap.preprocessing import format_input_data
from alphamap.sequenceplot import plot_peptide_traces, uniprot_color_dict, create_pdf_report, plot_3d_structure
from alphamap.uniprot_integration import uniprot_feature_dict
from alphamap.proteolytic_cleavage import protease_dict
from alphamap.organisms_data import all_organisms, import_fasta, import_uniprot_annotation


# LOCAL VARIABLES
full_fasta = None
full_uniprot = None
ac_gene_conversion = None
SETTINGS = {
    'max_file_size_gb': 50,
    'max_num_proteins_report': 100
}
SERVER = None
TAB_COUNTER = 0

# ERROR/WARNING MESSAGES
error_message_upload = "The selected {}file can't be uploaded. Please check the instructions for data uploading."
error_message_extract_samples = "It is impossible to extract raw file names from the file. Please check the list of necessary columns in the instructions for data uploading."
error_message_no_file = "The selected {}file is not found. Please check whether the specified path is correct."
error_message_upload_wrong_columns = "The columns necessary for further analysis cannot be extracted from the {} experimental file. Please check the data uploading instructions for a particular software tool."
error_message_size = f"A maximum file size shouldn't exceed {SETTINGS['max_file_size_gb']} GB."
error_message_report_long = f"Only first {SETTINGS['max_num_proteins_report']} proteins will be presented in the report."

if platform.system() == 'Windows':
    filepath_placeholder = 'D:\data\output_alphapept.csv'
else:
    filepath_placeholder = '/Users/data/output_alphapept.csv'

### PATHS
BASE_PATH = os.path.dirname(__file__)
DATA_PATH = os.path.join(BASE_PATH, "data")
IMAGE_PATH = os.path.join(BASE_PATH, "data", "img")

mpi_biochem_logo_path = os.path.join(IMAGE_PATH, "mpi_logo.png")
mpi_logo_path = os.path.join(IMAGE_PATH, "max-planck-gesellschaft.jpg")
github_logo_path = os.path.join(IMAGE_PATH, "github.png")

alphamap_tutorial_path = os.path.join(DATA_PATH, "alphamap_tutorial.pdf")
spectronaut_scheme_path = os.path.join(DATA_PATH, "spectronaut_export_scheme.rs")

uniprot_link_path = os.path.join(IMAGE_PATH, "uniprot_logo.png")
phosposite_link_path = os.path.join(IMAGE_PATH, "phosphosite_logo.png")
protter_link_path = os.path.join(IMAGE_PATH, "protter_logo.png")
pdb_link_path = os.path.join(IMAGE_PATH, "pdb_logo.png")
peptide_atlas_link_path = os.path.join(IMAGE_PATH, "peptide_atlas_logo.png")
alphafold_link_path = os.path.join(IMAGE_PATH, "alphamap_icon.png")

### Upload an extension and a GUI style
def get_css_style(
    file_name="gui_style.css",
    directory=DATA_PATH
):
    file = os.path.join(
        directory,
        file_name
    )
    with open(file) as f:
        return f.read()

def init_panel():
    pn.extension(raw_css=[get_css_style()])
    pn.extension('plotly')


### HEADER
header_titel = pn.pane.Markdown(
    '# AlphaMap',
    # width=1250,
    sizing_mode='stretch_width',
    css_classes=['main_header']
)
divider = pn.pane.HTML(
    '<hr style="height: 2px; border:none; background-color: #045082; width: 1480px">',
    # width=1500,
    sizing_mode='stretch_width',
    align='center'
)
mpi_biochem_logo = pn.pane.PNG(
    mpi_biochem_logo_path,
    link_url='https://www.biochem.mpg.de/mann',
    width=60,
    height=60,
    align='start'
)
mpi_logo = pn.pane.JPG(
    mpi_logo_path,
    link_url='https://www.biochem.mpg.de/en',
    height=62,
    embed=True,
    width=62,
    margin=(5, 0, 0, 5),
    css_classes=['opt']
)
github_logo = pn.pane.PNG(
    github_logo_path,
    link_url='https://github.com/MannLabs/pepmap',
    height=70,
    align='end'
)
header = pn.Row(
    mpi_biochem_logo,
    mpi_logo,
    header_titel,
    github_logo,
    height=70,
    sizing_mode='stretch_width'
)


### WIDGETS

#####################################
# SELECTORS
select_protein = pn.widgets.AutocompleteInput(
    name='Select a protein of interest:',
    placeholder='Type first letters ...',
    min_characters=1,
    case_sensitive=False
)
predefined_protein_list_titel = pn.pane.Markdown(
    'Load a list of pre-selected proteins:',
    margin=(0,0,0,12)
)
predefined_protein_list = pn.widgets.FileInput(
    accept=".txt",
    margin=(-10,0,5,12)
)
select_organism = pn.widgets.Select(
    name='Select an organism:',
    value='Human',
    options=['Human', 'Mouse', 'Rat', 'Cow', 'Zebrafish', 'Drosophila', 'Caenorhabditis elegans', 'Slime mold',
             'Arabidopsis thaliana', 'Rice', 'Escherichia coli', 'Bacillus subtilis', 'Saccharomyces cerevisiae',
             'SARS-CoV', 'SARS-CoV2'],
    align='center',
    margin=(0,0,0,7),
    width=300
)

#####################################
# RADIOBUTTONS/CHECKBOXES
search_by = pn.widgets.RadioBoxGroup(
    name='Search by',
    options=['Search by UniProt accession', 'Search by a gene name'],
    value='Search by UniProt accession'
)
select_all = pn.widgets.Checkbox(
    name='Select all',
    width=150
)
clear_all = pn.widgets.Checkbox(
    name='Clear all',
    width=150
)
proteases_select_all = pn.widgets.Checkbox(
    name='Select all',
    width=150
)
proteases_clear_all = pn.widgets.Checkbox(
    name='Clear all',
    width=150
)
alphafold_select_all = pn.widgets.Checkbox(
    name='Select all',
    width=150
)
alphafold_clear_all = pn.widgets.Checkbox(
    name='Clear all',
    width=150
)

#####################################
# RAW EXPERIMENTAL DATA

# first experimental file
experimental_data = pn.widgets.TextInput(
    name='Upload the first result file:',
    placeholder=filepath_placeholder,
    width=445,
    margin=(23,0,5,15)
)
experimental_data_loading_spinner = pn.indicators.LoadingSpinner(
    value=False,
    bgcolor='light',
    color='secondary',
    margin=(43,0,3,12),
    width=30,
    height=30
)
experimental_data_sample = pn.widgets.MultiSelect(
    name='Select samples:',
    disabled=True,
    size=5,
    width=770,
    margin=(0,0,20,14)
)
experimental_data_warning = pn.pane.Alert(
    width=550,
    height=30,
    alert_type="danger",
    margin=(-20,0,15,15)
)
experimental_data_sample_name = pn.widgets.TextInput(
    name='Sample name',
    disabled=True,
    width=130,
    margin=(23,0,5,11)
)
experimental_data_sample_name_remove_part = pn.widgets.TextInput(
    name='Prefix / suffix',
    disabled=True,
    width=130,
    margin=(23,0,5,10)
)
# second experimental file
experimental_data_2 = pn.widgets.TextInput(
    name='Upload the second result file:',
    placeholder=filepath_placeholder,
    width=445,
    margin=(23,0,5,15)
)
experimental_data_2_loading_spinner = pn.indicators.LoadingSpinner(
    value=False,
    bgcolor='light',
    color='secondary',
    margin=(43,0,3,12),
    width=30,
    height=30
)
experimental_data_2_sample = pn.widgets.MultiSelect(
    name='Select samples:',
    size=5,
    disabled=True,
    width=770,
    margin=(0,0,0,14)
)
experimental_data_2_warning = pn.pane.Alert(
    width=550,
    height=30,
    alert_type="danger",
    margin=(-20,0,15,15)
)
experimental_data_2_sample_name = pn.widgets.TextInput(
    name='Sample name',
    disabled=True,
    width=130,
    margin=(23,0,5,11)
)
experimental_data_2_sample_name_remove_part = pn.widgets.TextInput(
    name='Prefix / suffix',
    disabled=True,
    width=130,
    margin=(23,0,5,10)
)
# third experimental file
experimental_data_3 = pn.widgets.TextInput(
    name='Upload the third result file:',
    placeholder=filepath_placeholder,
    width=445,
    margin=(23,0,5,15)
)
experimental_data_3_loading_spinner = pn.indicators.LoadingSpinner(
    value=False,
    bgcolor='light',
    color='secondary',
    margin=(43,0,3,12),
    width=30,
    height=30
)
experimental_data_3_sample = pn.widgets.MultiSelect(
    name='Select samples:',
    size=5,
    disabled=True,
    width=770,
    margin=(0,0,20,14)
)
experimental_data_3_warning = pn.pane.Alert(
    width=550,
    height=30,
    alert_type="danger",
    margin=(-20,0,15,15)
)
experimental_data_3_sample_name = pn.widgets.TextInput(
    name='Sample name',
    disabled=True,
    width=130,
    margin=(23,0,5,11)
)
experimental_data_3_sample_name_remove_part = pn.widgets.TextInput(
    name='Prefix / suffix',
    disabled=True,
    width=130,
    margin=(23,0,5,10)
)
#####################################
# PREPROCESSED EXPERIMENTAL DATA
preprocessed_exp_data = pn.widgets.DataFrame(
    name='Exp_data'
)
preprocessed_exp_data_2 = pn.widgets.DataFrame(
    name='Exp_data_2'
)
preprocessed_exp_data_3 = pn.widgets.DataFrame(
    name='Exp_data_3'
)
upload_data_warning = pn.pane.Alert(
    width=900,
    alert_type="danger",
    margin=(-20, 50, 20, 50),
    align='center',
)
#####################################
# BUTTONS
upload_button = pn.widgets.Button(
    name='Upload data',
    button_type='primary',
    height=40,
    width=170,
    align='center',
)
upload_spinner = pn.indicators.LoadingSpinner(
    value=False,
    bgcolor='light',
    color='secondary',
    margin=(5,5,0,5),
    width=40,
    height=40
)

visualize_button = pn.widgets.Button(
    name='Visualize protein',
    button_type='primary',
    height=40,
    width=170,
    align='center',
    margin=(0,0,0,0)
)
visualize_spinner = pn.indicators.LoadingSpinner(
    value=False,
    bgcolor='light',
    color='secondary',
    align='center',
    margin=(0,0,20,20),
    width=40,
    height=40
)


def download_pdf_report():
    download_pdf_error.object = ''
    download_pdf_loading_spinner.value = True
    uniprot_options_combined = sum([each.value for each in uniprot_options.objects if each.value], [])
    # extract all experimental data and names
    all_data = []
    all_names = []
    if preprocessed_exp_data.value is not None:
        all_data.append(preprocessed_exp_data.value)
        all_names.append(
            extract_name(os.path.splitext(os.path.basename(experimental_data.value))[0],
                         experimental_data_sample.value,
                         experimental_data_sample_name.value,
                         experimental_data_sample_name_remove_part.value
            )
        )
    if preprocessed_exp_data_2.value is not None:
        all_data.append(preprocessed_exp_data_2.value)
        all_names.append(
            extract_name(os.path.splitext(os.path.basename(experimental_data_2.value))[0],
                         experimental_data_2_sample.value,
                         experimental_data_2_sample_name.value,
                         experimental_data_2_sample_name_remove_part.value
            )
        )
    if preprocessed_exp_data_3.value is not None:
        all_data.append(preprocessed_exp_data_3.value)
        all_names.append(
            extract_name(os.path.splitext(os.path.basename(experimental_data_3.value))[0],
                         experimental_data_3_sample.value,
                         experimental_data_3_sample_name.value,
                         experimental_data_3_sample_name_remove_part.value
            )
        )
    # if only one experimental file is uploaded we need to return a string for input into plot_peptide_traces
    if len(all_data) == 1:
        all_data = all_data[0]
        all_names = all_names[0]
    proteins_in_report = list(ac_gene_conversion.keys())
    if len(proteins_in_report) > SETTINGS['max_num_proteins_report']:
        download_pdf_error.object = error_message_report_long
        proteins_in_report = proteins_in_report[:SETTINGS['max_num_proteins_report']]
    report = create_pdf_report(
        proteins = proteins_in_report,
        df = all_data,
        name = all_names,
        fasta = full_fasta,
        uniprot = full_uniprot,
        selected_features=[uniprot_feature_dict[each] for each in uniprot_options_combined],
        uniprot_feature_dict=uniprot_feature_dict,
        uniprot_color_dict=uniprot_color_dict,
        selected_proteases=proteases_options.value,
        selected_alphafold_features=alphafold_options.value,
    )
    download_pdf_loading_spinner.value = False
    return report


download_pdf = pn.widgets.FileDownload(
    callback=download_pdf_report,
    label='PDF for a list of pre-selected proteins',
    disabled=True,
    filename='alphamap_pdf_report.pdf',
    button_type='default',
    height=31,
    width=390,
    margin=(5, 10, 0, 6),
)

download_pdf_loading_spinner = pn.indicators.LoadingSpinner(
    value=False,
    bgcolor='light',
    color='secondary',
    margin=(11,0,3,0),
    width=30,
    height=30
)

download_pdf_error = pn.pane.Alert(
    width=369,
    margin=(-22, 20, 20, 20),
    alert_type="primary",
    background='white'
)

### UNIPROT OPTIONS
options_preprocessing_events = pn.widgets.CheckButtonGroup(
    name='Molecule processing',
    value=['Chain', 'Initiator methionine', 'Peptide', 'Propeptide', 'Signal peptide', 'Transit peptide'],
    options=['Chain', 'Initiator methionine', 'Peptide', 'Propeptide', 'Signal peptide', 'Transit peptide'],
    align='center'
)
options_PTMs = pn.widgets.CheckButtonGroup(
    name='Post-translational modification',
    options=['Cross-link', 'Disulfide bond', 'Glycosylation', 'Lipidation', 'Modified residue'],
    value=['Cross-link', 'Disulfide bond', 'Glycosylation', 'Lipidation', 'Modified residue'],
    align='center'
)
options_domains = pn.widgets.CheckButtonGroup(
    name='Family & Domain',
    options=['Coiled coil', 'Compositional bias', 'Domain', 'Motif', 'Region', 'Repeat', 'Zinc finger'],
    value=['Coiled coil', 'Compositional bias', 'Domain', 'Motif', 'Region', 'Repeat', 'Zinc finger'],
    align='center'
)
options_locations = pn.widgets.CheckButtonGroup(
    name='Subcellular location',
    options=['Intramembrane', 'Topological domain', 'Transmembrane'],
    value=['Intramembrane', 'Topological domain', 'Transmembrane'],
    align='center'
)
options_functions = pn.widgets.CheckButtonGroup(
    name='Function',
    options=['Active site', 'Binding site', 'Calcium binding', 'DNA binding', 'Metal binding', 'Nucleotide binding', 'Site'],
    value=['Active site', 'Binding site', 'Calcium binding', 'DNA binding', 'Metal binding', 'Nucleotide binding', 'Site'],
    align='center'
)
options_sequences = pn.widgets.CheckButtonGroup(
    name='Sequence',
    options=['Alternative sequence', 'Natural variant', 'Non-adjacent residues', 'Non-standard residue',
             'Non-terminal residue', 'Sequence conflict', 'Sequence uncertainty'],
    value=['Alternative sequence', 'Natural variant', 'Non-adjacent residues', 'Non-standard residue',
             'Non-terminal residue', 'Sequence conflict', 'Sequence uncertainty'],
    align='center'
)
options_other = pn.widgets.CheckButtonGroup(
    name='Other options',
    options=['Secondary structure', 'Mutagenesis'],
    value=['Secondary structure', 'Mutagenesis'],
    align='center'
)

#####################################
### LAYOUTS
uniprot_options = pn.Accordion(
    options_preprocessing_events,
    options_PTMs,
    options_domains,
    options_locations,
    options_functions,
    options_sequences,
    options_other,
    active=list(range(0,7)),
    header_background='EAEAEA',
    active_header_background='EAEAEA',
    width = 850
)

uniprot_options_tab = pn.Card(
    uniprot_options,
    pn.Row(
        select_all,
        clear_all,
        margin=5
    ),
    title='UniProt annotations',
    collapsed=True,
    header_background='EAEAEA',
    active_header_background='EAEAEA',
    width=860,
    css_classes=['uniprot_options'],
)


### List of proteases
proteases_options = pn.widgets.CheckBoxGroup(
    options=list(protease_dict.keys()),
    value=['trypsin'],
    align='center',
    margin=10
)

custom_enzyme_field = pn.widgets.TextInput(
    name='Type a regular expression:',
    disabled=True,
    align='end',
    width=160,
    placeholder='([KR](?=[^P]))'
    # margin=(0,150,10,-30)S
)

proteases_options_tab = pn.Card(
    pn.Row(
        proteases_options,
        custom_enzyme_field
    ),
    pn.Row(
        proteases_select_all,
        proteases_clear_all,
        margin=5
    ),
    title='Protease cleavage sites',
    collapsed=True,
    header_background='EAEAEA',
    active_header_background='EAEAEA',
    width=860,
    css_classes=['uniprot_options'],
)


### List of alphafold options for linear sequence plot
alphafold_options = pn.widgets.CheckBoxGroup(
    options=["AlphaFold confidence", "AlphaFold exposure", "AlphaFold IDR", "AlphaFold secondary structures"],
    value=[''],
    align='center',
    margin=10
)

alphafold_options_tab = pn.Card(
    pn.Row(
        alphafold_options
    ),
    pn.Row(
        alphafold_select_all,
        alphafold_clear_all,
        margin=5
    ),
    title='AlphaFold annotation',
    collapsed=True,
    header_background='EAEAEA',
    active_header_background='EAEAEA',
    width=860,
    css_classes=['uniprot_options'],
)

### List of alphafold options for 3D sequence plot
alphafold3D_options = pn.widgets.Select(
    options=['AlphaFold confidence','AlphaFold exposure','AlphaFold IDR','AlphaFold secondary structures', 'MS peptides', 'MS modified peptides', 'MS PTMs', 'MS PTMs on AlphaFold IDR'],
    value='MS modified peptides',
    align='center',
    margin=10
)

alphafold3D_options_tab = pn.Card(
    pn.Row(
        alphafold3D_options
    ),
    title='Structure coloring',
    collapsed=False,
    header_background='EAEAEA',
    active_header_background='EAEAEA',
    width=860,
    css_classes=['uniprot_options'],
)

### MS data selection for 3D sequence plot
data3D_options = pn.widgets.Select(
    options=['all'],
    value='all',
    align='center',
    margin=10
)

@pn.depends(
    upload_button.param.clicks,
    watch=True
)
def update_data3D_options(clicks):
    if clicks > 0 and any(
        [experimental_data_sample.value, experimental_data_2_sample.value, experimental_data_3_sample.value]
    ):
        available_data_names = []
        if experimental_data.value:
            available_data_names.append(
                extract_name(os.path.splitext(os.path.basename(experimental_data.value))[0],
                             experimental_data_sample.value,
                             experimental_data_sample_name.value,
                             experimental_data_sample_name_remove_part.value
                )
            )
        if experimental_data_2.value:
            available_data_names.append(
                extract_name(os.path.splitext(os.path.basename(experimental_data_2.value))[0],
                             experimental_data_2_sample.value,
                             experimental_data_2_sample_name.value,
                             experimental_data_2_sample_name_remove_part.value
                )
            )
        if experimental_data_3.value:
            available_data_names.append(
                extract_name(os.path.splitext(os.path.basename(experimental_data_3.value))[0],
                             experimental_data_3_sample.value,
                             experimental_data_3_sample_name.value,
                             experimental_data_3_sample_name_remove_part.value
                )
            )
        if len(available_data_names) == 1:
            data3D_options.options = [available_data_names[0]]
            data3D_options.value = available_data_names[0]
        else:
            available_data_names.insert(0, "all")
            data3D_options.options = available_data_names
            data3D_options.value = available_data_names[0]

data3D_options_tab = pn.Card(
    pn.Row(
        data3D_options
    ),
    title='MS data selection',
    collapsed=False,
    header_background='EAEAEA',
    active_header_background='EAEAEA',
    width=860,
    css_classes=['uniprot_options'],
)

### MAIN PART
project_description = pn.pane.Markdown(
    """### AlphaMap enables the exploration of proteomic datasets on the peptide level. It is possible to evaluate the sequence coverage of any identified protein and its post-translational modifications (PTMs). AlphaMap further integrates all available UniProt sequence annotations as well as information about proteolytic cleavage sites.""",
    margin=(0, 0, -20, 0),
    css_classes=['main-part'],
    width=700
)

divider_descr = pn.pane.HTML(
    '<hr style="height: 8px; border:none; background-color: #045082; width: 640px">',
    # width=1510,
    sizing_mode='stretch_width',
    align='center'
)

project_instuction = pn.pane.Markdown(
    """#### How to use AlphaMap:
    1. Select the organism of your proteomic study.
    2. Provide the filepath to your proteomic datasets analyzed by
    AlphaPept, MaxQuant, Spectronaut, DIA-NN or FragPipe.
        - Wait for samples to be displayed in the 'Select samples' field.
        - (optional) Select either all samples (default) or any specific
        sample(s) to visualize together as one trace.
        - (optional) Choose a name by which the selected sample(s) will
        be displayed in the figure.
        - (optional) Provide a prefix or suffix to be removed from the
        original names of the selected samples.
        * Up to three datasets or sets of selected samples can be
        visualized together.
    3. Press the 'Upload Data' button.
    4. Select a protein of interest by UniProt accession or gene name.
    5. (optional) Load a list of pre-selected proteins to reduce the list
    of available proteins.
    6. Select annotation options for either a linear sequence plot or
    a 3D visualization based on structure predictions from AlphaFold2.
    7. Press the 'Visualize Protein' button.
    8. Enjoy exploring your data!
    """,
    width=530,
    align='start',
    margin=(0, 80, 0, 10)
)

alphamap_tutorial = pn.widgets.FileDownload(
    file=alphamap_tutorial_path,
    label='AlphaMap tutorial',
    button_type='default',
    auto=True,
    width=530,
    align='start',
    margin=(10, 80, 5, 10),
    css_classes=['spectronaut_instr']
)

spectronaut_description = pn.pane.Markdown(
    """
    The data needs to be exported in the **normal long** format as .tsv or .csv file.

    It needs to include the following columns:
    >- PEP.AllOccurringProteinAccessions
    >- EG.ModifiedSequence
    >- R.FileName

    To ensure the correct export format from Spectronaut, you can download and apply the provided export scheme “spectronaut_export_scheme.rs”.
    """,
    width=530,
    align='start',
    margin=(0, 80, 0, 20)
)

spectronaut_scheme = pn.widgets.FileDownload(
    file=spectronaut_scheme_path,
    filename='spectronaut_export_scheme.rs',
    button_type='default',
    auto=True,
    css_classes=['button_options'],
)

maxquant_description = pn.pane.Markdown(
    """
    To visualize the proteins which were analyzed by the MaxQuant software please use the **evidence.txt** file.

    The following columns from the file are used for visualization:
    >- Proteins
    >- Modified sequence
    >- Raw file
    """,
    width=530,
    align='start',
    margin=(0, 80, 0, 20)
)

alphapept_description = pn.pane.Markdown(
    """
    To visualize the proteins which were analyzed by the AlphaPept software please use the **results.csv** file.

    The following columns from the file are used for visualization:
    >- protein_group
    >- sequence
    >- shortname
    """,
    width=530,
    align='start',
    margin=(0, 80, 0, 20)
)

diann_description = pn.pane.Markdown(
    """
    To visualize the proteins which were analyzed by the DIA-NN software please use the **{experiment_name}.tsv** file.

    The following columns from the file are used for visualization:
    >- Protein.Ids
    >- Modified.Sequence
    >- Run
    """,
    width=530,
    align='start',
    margin=(0, 80, 0, 20)
)

fragpipe_description = pn.pane.Markdown(
    """
    There are two options to visualize data analyzed by FragPipe:

    1) Upload individual **"peptide.tsv"** files for single MS runs. In this case, the following columns from the original file are used for visualization:
    >- Protein ID
    >- Peptide
    >- Assigned Modifications

    2) Upload the **"combined_peptide.tsv"** file with the joint information about peptides identified in all runs (there is an option to select the experiment(s)). Be aware that the combined_peptide.tsv does not provide information about PTM localization. PTMs are therefore not shown for this option. Following columns are used for visalization:
    >- Protein ID
    >- Sequence
    >- All 'Spectral Count' columns containing information about individual experiments
    """,
    width=530,
    align='start',
    margin=(0, 80, 0, 20)
)

spectronaut_instructions = pn.Card(
    spectronaut_description,
    spectronaut_scheme,
    title='Spectronaut instructions',
    collapsed=True,
    width=530,
    align='start',
    margin=(0, 80, 5, 10),
    css_classes=['spectronaut_instr']
)

maxquant_instructions = pn.Card(
    maxquant_description,
    title='MaxQuant instructions',
    collapsed=True,
    width=530,
    align='start',
    margin=(0, 80, 5, 10),
    css_classes=['spectronaut_instr']
)

alphapept_instructions = pn.Card(
    alphapept_description,
    title='AlphaPept instructions',
    collapsed=True,
    width=530,
    align='start',
    margin=(0, 80, 5, 10),
    css_classes=['spectronaut_instr']
)

diann_instructions = pn.Card(
    diann_description,
    title='DIA-NN instructions',
    collapsed=True,
    width=530,
    align='start',
    margin=(0, 80, 5, 10),
    css_classes=['spectronaut_instr']
)

fragpipe_instructions = pn.Card(
    fragpipe_description,
    title='FragPipe instructions',
    collapsed=True,
    width=530,
    align='start',
    margin=(0, 80, 5, 10),
    css_classes=['spectronaut_instr']
)

additional_data_card = pn.Card(
    pn.Row(
        experimental_data_2,
        experimental_data_2_loading_spinner,
        experimental_data_2_sample_name,
        experimental_data_2_sample_name_remove_part
    ),
    experimental_data_2_warning,
    experimental_data_2_sample,
    pn.Row(
        experimental_data_3,
        experimental_data_3_loading_spinner,
        experimental_data_3_sample_name,
        experimental_data_3_sample_name_remove_part
    ),
    experimental_data_3_warning,
    experimental_data_3_sample,
    title='Upload additional result files',
    collapsed=True,
    width=780,
    margin=(2,0,0,10),
    css_classes=['add_experim_options']
)

selection_box = pn.Column(
    select_organism,
    pn.Row(
        experimental_data,
        experimental_data_loading_spinner,
        experimental_data_sample_name,
        experimental_data_sample_name_remove_part
    ),
    experimental_data_warning,
    experimental_data_sample,
    additional_data_card,
    margin=(0, 30, 10, 30),
    width=790,
    css_classes=['selection_box'],
)


main_part = pn.Column(
    project_description,
    divider_descr,
    pn.Row(
        pn.Column(
            project_instuction,
            alphamap_tutorial,
            spectronaut_instructions,
            maxquant_instructions,
            alphapept_instructions,
            diann_instructions,
            fragpipe_instructions
        ),
        selection_box,
        align='center',
        sizing_mode='stretch_width',
    ),
    pn.Row(
        upload_button,
        upload_spinner,
        align='center',
        margin=(0, 0),
        sizing_mode='stretch_width',
    ),
    pn.Row(
        upload_data_warning,
        align='center',
        width=1000,
        height=60
    ),
    background='#eaeaea',
    sizing_mode='stretch_width',
    margin=(5, 0, 60, 0)
)


# switch to different websites
uniprot_link = pn.pane.PNG(
    uniprot_link_path,
    width=120,
    height=60,
    # align='center',
    margin=(0, 30, 0, 40)
)
phosposite_link = pn.pane.PNG(
    phosposite_link_path,
    width=200,
    height=60,
    # align='center',
    margin=(0, 20)
)
protter_link = pn.pane.PNG(
    protter_link_path,
    width=140,
    height=60,
    # align='center',
    margin=(0, 20)
)
pdb_link = pn.pane.PNG(
    pdb_link_path,
    width=120,
    height=60,
    # align='center',
    margin=(0, 20)
)
peptide_atlas_link = pn.pane.PNG(
    peptide_atlas_link_path,
    width=120,
    height=60,
    # align='center',
    margin=(0, 20)
)
alphafold_link = pn.pane.PNG(
    alphafold_link_path,
    width=180,
    height=60,
    # align='center',
    margin=(0, 20)
)


def extract_uniprot_ai(protein, search_by):
    if search_by == 'Search by a gene name':
        return re.findall(r"\((?P<id>.+?)\)", protein)[0]
    return protein


def update_all_links():
    selected_protein_id = extract_uniprot_ai(select_protein.value, search_by.value)
    uniprot_link.link_url = 'https://www.uniprot.org/uniprot/' + selected_protein_id
    phosposite_link.link_url = 'http://www.phosphosite.org/uniprotAccAction?id=' + selected_protein_id
    protter_link.link_url = 'https://wlab.ethz.ch/protter/#up=' + selected_protein_id
    pdb_link.link_url = f'https://www.rcsb.org/search?request=%7B%22query%22%3A%7B%22parameters%22%3A%7B%22value%22%3A%22{selected_protein_id}%22%7D%2C%22type%22%3A%22terminal%22%2C%22service%22%3A%22text%22%2C%22node_id%22%3A0%7D%2C%22return_type%22%3A%22entry%22%2C%22request_options%22%3A%7B%22pager%22%3A%7B%22start%22%3A0%2C%22rows%22%3A100%7D%2C%22scoring_strategy%22%3A%22combined%22%2C%22sort%22%3A%5B%7B%22sort_by%22%3A%22score%22%2C%22direction%22%3A%22desc%22%7D%5D%7D%2C%22request_info%22%3A%7B%22src%22%3A%22ui%22%2C%22query_id%22%3A%223407f72e3370cd10196490437be3ec87%22%7D%7D'
    peptide_atlas_link.link_url = f"https://db.systemsbiology.net/sbeams/cgi/PeptideAtlas/GetProtein?protein_name={selected_protein_id}&action=QUERY"
    alphafold_link.link_url = 'https://alphafold.ebi.ac.uk/entry/' + selected_protein_id

def visualize_buttons():
    if select_protein.value:
        update_all_links()
        buttons_layout = pn.Row(
            pn.pane.Markdown(
                "### Inspect target protein on other platforms:",
                margin=(10, 0, 0, 50)
            ),
            pn.layout.VSpacer(width=30),
            uniprot_link,
            phosposite_link,
            protter_link,
            pdb_link,
            peptide_atlas_link,
            alphafold_link,
            height=60,
            sizing_mode='stretch_width',
            margin=(50,0)
        )
        return buttons_layout
    else:
        return None


### PREPROCESSING
def upload_experimental_data():
    global ac_gene_conversion
    all_unique_proteins = []
    preprocessed_exp_data.value = preprocessed_exp_data_2.value = preprocessed_exp_data_3.value = None
    upload_data_warning.object = ""
    if experimental_data.value:
        if experimental_data_sample.value == ['All samples']:
            data_samples = None
        else:
            data_samples = experimental_data_sample.value
        try:
            preprocessed_exp_data.value = format_input_data(
                df = import_data(
                    experimental_data.value.replace("\\", "/").replace('"', ''),
                    verbose=False,
                    sample=data_samples
                ),
                fasta = full_fasta,
                modification_exp = r'\[.*?\]',
                verbose = False)
            all_unique_proteins.extend(preprocessed_exp_data.value.unique_protein_id.unique().tolist())
        except (TypeError, MemoryError, FileNotFoundError, ValueError, AttributeError) as e:
            if type(e).__name__ == 'MemoryError':
                upload_data_warning.object = error_message_size
            elif type(e).__name__  in ['TypeError', 'ValueError', 'AttributeError']:
                upload_data_warning.object = error_message_upload_wrong_columns.format('first')
            elif type(e).__name__ == 'FileNotFoundError':
                upload_data_warning.object = error_message_no_file.format('first experimental ')
            upload_spinner.value = False
    if experimental_data_2.value:
        if experimental_data_2_sample.value == ['All samples']:
            data_2_samples = None
        else:
            data_2_samples = experimental_data_2_sample.value
        try:
            preprocessed_exp_data_2.value = format_input_data(
                df = import_data(
                    experimental_data_2.value.replace("\\", "/").replace('"', ''),
                    verbose=False,
                    sample=data_2_samples
                ),
                fasta = full_fasta,
                modification_exp = r'\[.*?\]',
                verbose = False)
            all_unique_proteins.extend(preprocessed_exp_data_2.value.unique_protein_id.unique().tolist())
        except (TypeError, MemoryError, FileNotFoundError, ValueError,
        AttributeError) as e:
            if not upload_data_warning.object:
                if type(e).__name__ == 'MemoryError':
                    upload_data_warning.object = error_message_size
                elif type(e).__name__  in ['TypeError', 'ValueError', 'AttributeError']:
                    upload_data_warning.object = error_message_upload_wrong_columns.format('second')
                elif type(e).__name__ == 'FileNotFoundError':
                    upload_data_warning.object = error_message_no_file.format('second experimental ')
            upload_spinner.value = False
    if experimental_data_3.value:
        if experimental_data_3_sample.value == ['All samples']:
            data_3_samples = None
        else:
            data_3_samples = experimental_data_3_sample.value
        try:
            preprocessed_exp_data_3.value = format_input_data(
                df = import_data(
                    experimental_data_3.value.replace("\\", "/").replace('"', ''),
                    verbose=False,
                    sample=data_3_samples
                ),
                fasta = full_fasta,
                modification_exp = r'\[.*?\]',
                verbose = False)
            all_unique_proteins.extend(preprocessed_exp_data_3.value.unique_protein_id.unique().tolist())
        except (TypeError, MemoryError, FileNotFoundError, ValueError, AttributeError) as e:
            if not upload_data_warning.object:
                if type(e).__name__ == 'MemoryError':
                    upload_data_warning.object = error_message_size
                elif type(e).__name__  in ['TypeError', 'ValueError', 'AttributeError']:
                    upload_data_warning.object = error_message_upload_wrong_columns.format('third')
                elif type(e).__name__ == 'FileNotFoundError':
                    upload_data_warning.object = error_message_no_file.format('third experimental ')
            upload_spinner.value = False
    ac_gene_conversion = {
        each: f"{full_fasta.get_by_id(each).description.get('GN')} ({full_fasta.get_by_id(each).description.get('id')})" \
        for each in sorted(list(set(all_unique_proteins)))}
    # to set a selection list of availible proteins depending which user wants to search by
    if search_by.value == 'Search by a gene name':
        select_protein.options = list(ac_gene_conversion.values())
    else:
        select_protein.options = list(ac_gene_conversion.keys())


@pn.depends(
    predefined_protein_list.param.value,
    watch=True
)
def filter_proteins(data):
    if data:
        download_pdf.disabled=False
        global ac_gene_conversion
        predefined_list = []
        for line in StringIO(str(data, "utf-8")).readlines():
            predefined_list.append(line.strip().upper())
        ac_gene_conversion = {k:v for k,v in ac_gene_conversion.items() if (k in predefined_list or v.split()[0] in predefined_list)}
        if search_by.value == 'Search by a gene name':
            select_protein.options = list(ac_gene_conversion.values())
        else:
            select_protein.options = list(ac_gene_conversion.keys())


def upload_organism_info():
    global full_fasta, full_uniprot

    full_fasta = import_fasta(select_organism.value)
    full_uniprot = import_uniprot_annotation(select_organism.value)


def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
    return sorted(l, key = alphanum_key)


def extract_samples(path):
    """
    Extract information about unique sample names that present in the raw file analyzed by MaxQuant, Spectronaut, AlphaPept or DIA-NN.
    """
    path = path.replace('"', '')
    file_size_gb = os.stat(path).st_size / 1024**3
    if file_size_gb > SETTINGS['max_file_size_gb']:
        raise MemoryError
    # try:
    unique_samples = extract_rawfile_unique_values(path.replace("\\", "/"))
    # except:
    #     raise TypeError("This file can't be uploaded.")
    unique_samples_sorted = natural_sort(unique_samples)
    return unique_samples_sorted


def extract_name(filename, sample, sample_name, sample_name_remove_prefix):
    if sample_name:
        name = sample_name
    elif sample and sample != ['All samples']:
        if isinstance(sample, list):
            name = ";".join(sample).replace(sample_name_remove_prefix, '')
        else:
            name = sample.replace(sample_name_remove_prefix, '')
    else:
        name = filename.split('.')[0]
    return name


@pn.depends(
    experimental_data_sample_name.param.value,
    experimental_data_2_sample_name.param.value,
    experimental_data_3_sample_name.param.value,
    experimental_data_sample_name_remove_part.param.value,
    experimental_data_2_sample_name_remove_part.param.value,
    experimental_data_3_sample_name_remove_part.param.value,
    experimental_data_sample.param.value,
    experimental_data_2_sample.param.value,
    experimental_data_3_sample.param.value,
    experimental_data.param.value,
    experimental_data_2.param.value,
    experimental_data_3.param.value,
    select_organism.param.value,
    watch=True
)
def clear_dashboard(*args):
    global download_pdf
    download_pdf = pn.widgets.FileDownload(
        callback=download_pdf_report,
        label='PDF for a list of pre-selected proteins',
        filename='alphamap_pdf_report.pdf',
        disabled=True,
        button_type='default',
        height=31,
        width=369,
        margin=(5, 20, 15, 12),
        align='center'
    )
    upload_button.clicks = 0
    visualize_button.clicks = 0
    predefined_protein_list.value = None
    download_pdf_error.object = ''
    upload_data
    visualize_plot


@pn.depends(
    experimental_data.param.value,
    watch=True
)
def update_data_sample_info(data1):
    experimental_data_loading_spinner.value = True
    experimental_data_sample.disabled = False
    try:
        experimental_data_warning.object = None
        experimental_data_sample.options = ['All samples'] + extract_samples(data1)
        experimental_data_sample.value = ['All samples']
        experimental_data_sample_name_remove_part.disabled = False
    except (TypeError, MemoryError, FileNotFoundError, ValueError) as e:
        if type(e).__name__ == 'MemoryError':
            experimental_data_warning.object = error_message_size
        elif type(e).__name__ == 'TypeError':
            experimental_data_warning.object = error_message_upload.format('')
        elif type(e).__name__ == 'ValueError':
            experimental_data_warning.object = error_message_extract_samples
        elif type(e).__name__ == 'FileNotFoundError':
            if data1 == "":
                experimental_data_warning.object = ""
            else:
                experimental_data_warning.object = error_message_no_file.format('')
        preprocessed_exp_data.value = None
        experimental_data_sample.disabled = True
        experimental_data_sample_name.disabled = True
        experimental_data_sample_name.value = ''
        experimental_data_sample_name_remove_part.disabled = True
        experimental_data_sample_name_remove_part.value = ''
        experimental_data_sample.options = []
        experimental_data_sample.value = []
    experimental_data_loading_spinner.value = False


@pn.depends(
    experimental_data_2.param.value,
    watch=True
)
def update_data_2_sample_info(data2):
    experimental_data_2_loading_spinner.value = True
    experimental_data_2_sample.disabled = False
    try:
        experimental_data_2_warning.object = None
        experimental_data_2_sample.options = ['All samples'] + extract_samples(data2)
        experimental_data_2_sample.value = ['All samples']
        experimental_data_2_sample_name_remove_part.disabled = False
    except (TypeError, MemoryError, FileNotFoundError, ValueError) as e:
        if type(e).__name__ == 'MemoryError':
            experimental_data_2_warning.object = error_message_size
        elif type(e).__name__ == 'TypeError':
            experimental_data_2_warning.object = error_message_upload.format('')
        elif type(e).__name__ == 'ValueError':
            experimental_data_2_warning.object = error_message_extract_samples
        elif type(e).__name__ == 'FileNotFoundError':
            if data2 == "":
                experimental_data_2_warning.object = ""
            else:
                experimental_data_2_warning.object = error_message_no_file.format('')
        preprocessed_exp_data_2.value = None
        experimental_data_2_sample.disabled = True
        experimental_data_2_sample_name.disabled = True
        experimental_data_2_sample_name.value = ''
        experimental_data_2_sample_name_remove_part.disabled = True
        experimental_data_2_sample_name_remove_part.value = ''
        experimental_data_2_sample.options = []
        experimental_data_2_sample.value = []
    experimental_data_2_loading_spinner.value = False


@pn.depends(
    experimental_data_3.param.value,
    watch=True
)
def update_data_3_sample_info(data3):
    experimental_data_3_loading_spinner.value = True
    experimental_data_3_sample.disabled = False
    try:
        experimental_data_3_warning.object = None
        experimental_data_3_sample.options = ['All samples'] + extract_samples(data3)
        experimental_data_3_sample.value = ['All samples']
        experimental_data_3_sample_name_remove_part.disabled = False
    except (TypeError, MemoryError, FileNotFoundError, ValueError) as e:
        if type(e).__name__ == 'MemoryError':
            experimental_data_3_warning.object = error_message_size
        elif type(e).__name__ == 'TypeError':
            experimental_data_3_warning.object = error_message_upload.format('')
        elif type(e).__name__ == 'ValueError':
            experimental_data_3_warning.object = error_message_extract_samples
        elif type(e).__name__ == 'FileNotFoundError':
            if data3 == "":
                experimental_data_3_warning.object = ""
            else:
                experimental_data_3_warning.object = error_message_no_file.format('')
        preprocessed_exp_data_3.value = None
        experimental_data_3_sample.disabled = True
        experimental_data_3_sample_name.disabled = True
        experimental_data_3_sample_name.value = ''
        experimental_data_3_sample_name_remove_part.disabled = True
        experimental_data_3_sample_name_remove_part.value = ''
        experimental_data_3_sample.options = []
        experimental_data_3_sample.value = []
    experimental_data_3_loading_spinner.value = False


@pn.depends(
    experimental_data_sample.param.value,
    experimental_data_2_sample.param.value,
    experimental_data_3_sample.param.value,
    watch=True
)
def change_sample_name_state(data_sample, data_2_sample, data_3_sample):
    if data_sample:
        experimental_data_sample_name.disabled = False
        experimental_data_sample_name_remove_part.disabled = False
    else:
        experimental_data_sample_name.disabled = True
        experimental_data_sample_name_remove_part.disabled = True
    if data_2_sample:
        experimental_data_2_sample_name.disabled = False
        experimental_data_2_sample_name_remove_part.disabled = False
    else:
        experimental_data_2_sample_name.disabled = True
        experimental_data_2_sample_name_remove_part.disabled = True
    if data_3_sample:
        experimental_data_3_sample_name.disabled = False
        experimental_data_3_sample_name_remove_part.disabled = False
    else:
        experimental_data_3_sample_name.disabled = True
        experimental_data_3_sample_name_remove_part.disabled = True


@pn.depends(
    select_all.param.value,
    clear_all.param.value,
    watch=True
)
def change_uniprot_selection(select, clear):
    if select:
        clear_all.value = False
        select_all.value = False
        for each in uniprot_options.objects:
            each.value = each.options
    if clear:
        clear_all.value = False
        select_all.value = False
        for each in uniprot_options.objects:
            each.value = []


@pn.depends(
    proteases_select_all.param.value,
    proteases_clear_all.param.value,
    watch=True
)
def change_proteases_selection(select, clear):
    if select:
        proteases_clear_all.value = False
        proteases_select_all.value = False
        proteases_options.value = proteases_options.options
    if clear:
        proteases_clear_all.value = False
        proteases_select_all.value = False
        proteases_options.value = []


@pn.depends(
    custom_enzyme_field.param.value,
    watch=True
)
def update_protease_dict(_):
    protease_dict['custom_enzyme'] = custom_enzyme_field.value


@pn.depends(
    proteases_options.param.value,
    watch=True
)
def de_activate_custom_enzyme_field(_):
    if 'custom_enzyme' in proteases_options.value:
        custom_enzyme_field.disabled = False
    else:
        custom_enzyme_field.disabled = True

@pn.depends(
    alphafold_select_all.param.value,
    alphafold_clear_all.param.value,
    watch=True
)
def change_alphafold_selection(select, clear):
    if select:
        alphafold_clear_all.value = False
        alphafold_select_all.value = False
        alphafold_options.value = alphafold_options.options
    if clear:
        alphafold_clear_all.value = False
        alphafold_select_all.value = False
        alphafold_options.value = []

@pn.depends(
    search_by.param.value,
    watch=True
)
def change_autocomplete_input(search_by):
    if any(
        [experimental_data_sample.value, experimental_data_2_sample.value, experimental_data_3_sample.value]
    ):
        if search_by == 'Search by a gene name':
            select_protein.options = list(ac_gene_conversion.values())
        else:
            select_protein.options = list(ac_gene_conversion.keys())


### Tabs for linear and 3D plots
plot_selection_tabs = pn.Tabs(
    ('Linear sequence plot', pn.Column(uniprot_options_tab,proteases_options_tab,alphafold_options_tab)),
    ('3D plot', pn.Column(alphafold3D_options_tab, data3D_options_tab)))

### VISUALIZATION
import alphamap
import os
import tempfile
BASE_PATH = os.path.dirname(alphamap.__file__)
js_path_global = os.path.join(BASE_PATH, 'js')
cif_path_global = tempfile.gettempdir()
pn.extension(js_files={'js': 'js/alphafold-viz.js'})
pn.extension(css_files=['js/alphafold-viz.css'])

@pn.depends(
    upload_button.param.clicks
)
def upload_data(clicks):
    if clicks > 0 and any(
        [experimental_data_sample.value, experimental_data_2_sample.value, experimental_data_3_sample.value]
    ):
        upload_spinner.value = True
        select_protein.value = None
        # preload the data
        upload_organism_info()
        upload_experimental_data()
        # create a layout
        if len(upload_data_warning.object) == 0:
            app = pn.Column(
                pn.Row(
                    pn.Column(
                        select_protein,
                        search_by,
                        predefined_protein_list_titel,
                        predefined_protein_list,
                        pn.Row(
                            download_pdf,
                            download_pdf_loading_spinner
                        ),
                        download_pdf_error
                    ),
                    pn.layout.VSpacer(width=80),
                    pn.Column(plot_selection_tabs),
                    align='center'
                ),
                pn.layout.HSpacer(height=4),
                pn.Row(
                    visualize_button,
                    visualize_spinner,
                    align='center'
                ),
                divider,
                align='center',
                sizing_mode='stretch_width',
                margin=(20, 0)
            )
            upload_spinner.value = False
            return app
        else:
            return None
    else:
        return None

@pn.depends(
    visualize_button.param.clicks
)
def visualize_plot(clicks):
    if select_protein.value and clicks > 0:
        visualize_spinner.value = True
        # combine selected uniprot options in one list
        uniprot_options_combined = sum([each.value for each in uniprot_options.objects if each.value], [])
        # extract all experimental data and names
        all_data = []
        all_names = []
        if preprocessed_exp_data.value is not None:
            all_data.append(preprocessed_exp_data.value)
            all_names.append(
                extract_name(os.path.splitext(os.path.basename(experimental_data.value))[0],
                             experimental_data_sample.value,
                             experimental_data_sample_name.value,
                             experimental_data_sample_name_remove_part.value
                )
            )
        if preprocessed_exp_data_2.value is not None:
            all_data.append(preprocessed_exp_data_2.value)
            all_names.append(
                extract_name(os.path.splitext(os.path.basename(experimental_data_2.value))[0],
                             experimental_data_2_sample.value,
                             experimental_data_2_sample_name.value,
                             experimental_data_2_sample_name_remove_part.value
                )
            )
        if preprocessed_exp_data_3.value is not None:
            all_data.append(preprocessed_exp_data_3.value)
            all_names.append(
                extract_name(os.path.splitext(os.path.basename(experimental_data_3.value))[0],
                             experimental_data_3_sample.value,
                             experimental_data_3_sample_name.value,
                             experimental_data_3_sample_name_remove_part.value
                )
            )
        # if only one experimental file is uploaded we need to return a string for input into plot_peptide_traces
        if len(all_data) == 1:
            all_data = all_data[0]
            all_names = all_names[0]
        try:
            if search_by.value == 'Search by a gene name':
                selected_protein = re.findall(r"\((?P<id>.+?)\)", select_protein.value)[0]
            else:
                selected_protein = select_protein.value
        except IndexError:
            visualize_spinner.value = False
            return None
        # create a main figure
        if plot_selection_tabs.active == 0:
            fig =  plot_peptide_traces(
                df = all_data,
                name = all_names,
                protein = selected_protein,
                fasta = full_fasta,
                uniprot = full_uniprot,
                selected_features = [uniprot_feature_dict[each] for each in uniprot_options_combined],
                uniprot_feature_dict = uniprot_feature_dict,
                uniprot_color_dict = uniprot_color_dict,
                selected_proteases=proteases_options.value,
                selected_alphafold_features=alphafold_options.value,
                dashboard=True
            )
            plot =  pn.Column(
                pn.Pane(
                    fig,
                    config={'toImageButtonOptions':
                               {'format': 'svg', # one of png, svg, jpeg, webp
                                'filename': f"alphamap_{full_fasta[selected_protein].description['name']}_{full_fasta[selected_protein].description['id']}",
                                'height': 500,
                                'width': 1500,
                                'scale': 1 # Multiply title/legend/axis/canvas sizes by this factor
                               }
                           },
                    align='center',
                    sizing_mode='stretch_width',
                    # width=1500
                ),
                visualize_buttons,
                align='center',
                sizing_mode='stretch_width'
            )
        else:
            if not ((data3D_options.value == 'all') or (isinstance(all_data, pd.DataFrame))):
                selected_data_idx = all_names.index(data3D_options.value)
                all_data = all_data[selected_data_idx]

            print(all_data)

            plot3D_html, js_path, cif_path = plot_3d_structure(
                    df = all_data,
                    name = data3D_options.value,
                    protein = selected_protein,
                    fasta = full_fasta,
                    selected_coloring = alphafold3D_options.value,
                    dashboard = True)

            plot =  pn.Row(pn.Column(pn.layout.HSpacer(height=80),
                            pn.panel(plot3D_html, width = 1500, align='center', sizing_mode='stretch_width'),
                            visualize_buttons,
                            align='center', sizing_mode='stretch_width'),
                            align='center', sizing_mode='stretch_width')

            global js_path_global
            js_path_global = js_path
            global cif_path_global
            cif_path_global = cif_path

        visualize_spinner.value = False
        return plot
    else:
        return None


def run():
    import alphamap
    import bokeh.server.views.ws
    global SERVER

    init_panel()
    layout = pn.Column(
        header,
        main_part,
        upload_data,
        visualize_plot,
        sizing_mode='stretch_width'
    )
    original_open = bokeh.server.views.ws.WSHandler.open
    bokeh.server.views.ws.WSHandler.open = open_browser_tab(original_open)
    original_on_close = bokeh.server.views.ws.WSHandler.on_close
    bokeh.server.views.ws.WSHandler.on_close = close_browser_tab(
        original_on_close
    )
    print("*"*30)
    print(f"* AlphaMap {alphamap.__version__} *".center(30, '*'))
    print("*"*30)
    #SERVER = layout.show(threaded=True, title='AlphaMap')
    #SERVER.join()
    SERVER = pn.serve(layout, threaded=True, title='AlphaMap', static_dirs={'js': js_path_global, 'cif': cif_path_global})


def open_browser_tab(func):
    def wrapper(*args, **kwargs):
        global TAB_COUNTER
        TAB_COUNTER += 1
        return func(*args, **kwargs)
    return wrapper


def close_browser_tab(func):
    def wrapper(*args, **kwargs):
        global TAB_COUNTER
        TAB_COUNTER -= 1
        return_value = func(*args, **kwargs)
        if TAB_COUNTER == 0:
            quit_server()
        return return_value
    return wrapper


def quit_server():
    print("Quitting server...")
    SERVER.stop()


### JS callbacks to control the behaviour of pn.Cards
uniprot_options_tab.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.uniprot_options');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': uniprot_options_tab}
);

proteases_options_tab.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.uniprot_options');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': proteases_options_tab}
);

alphafold_options_tab.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.uniprot_options');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': alphafold_options_tab}
);

alphafold3D_options_tab.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.uniprot_options');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': alphafold3D_options_tab}
);

data3D_options_tab.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.uniprot_options');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': data3D_options_tab}
);

additional_data_card.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.add_experim_options');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': additional_data_card}
);

spectronaut_instructions.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.spectronaut_instr');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': spectronaut_instructions}
);

maxquant_instructions.jscallback(
    collapsed="""
        var $container = $("html,body");
        var $scrollTo = $('.spectronaut_instr');

        $container.animate({scrollTop: $container.offset().top + $container.scrollTop(), scrollLeft: 0},300);
        """,
    args={'card': maxquant_instructions}
);


if __name__ == '__main__':
    run()
